package org.softwarepipelines.config;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;

import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.pipeline.PipelineConfig;
import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.softwarepipelines.config.util.ConfigHelper;

import org.w3c.dom.Element;

/**
 * PipelinesConfig
 * 
 * Loads the pipelines configuration from XML.
 *
 * @author cholm
 */
public class PipelinesMasterConfig {
	final private static Logger LOG = Logger.getLogger( PipelinesMasterConfig.class );
    
    /* Config File */
    protected File configFile;
    
    /* Instrumentor Log Interval */
    protected int instrumentorLogInterval;
    
    /* Routers */
    protected List<PipelinesDistributorConfig> distributors;
    
    public PipelinesMasterConfig() {
    }
    
    public PipelinesMasterConfig(final File conf) throws ConfigException, IOException {
    	// set the configuration file
    	this.configFile = conf;
    	
    	// and load the configuration
    	this.load();	
    }
    
    public void load() throws ConfigException, IOException {
    	DynamicConfig dc;
    	LOG.debug("Loading configuration from file: " + configFile.getAbsolutePath());
    	
    	// first, instantiate the distributors list
    	distributors = new ArrayList<PipelinesDistributorConfig>();

    	// get the DynamicConfig instance from the factory
    	dc = DynamicConfigFactory.load(configFile);
    		
    	// and then load the distributors
        loadDistributors(dc);
        
        // finally, set the relay service connectors (now that all the connector configs are complete)
        setupRelays();
    }
    
    private void loadDistributors(final DynamicConfig dc) throws ConfigException {
    	// get the instrumentor element
    	String xPath = ConfigHelper.INSTRUMENTOR_ELEMENT;
    	List<Element> elements = dc.getElements(xPath);
    	this.setInstrumentorLogInterval(Integer.parseInt(elements.get(0).getAttribute(ConfigHelper.LOG_INT_ATTR)));
    	
    	// get the list of pipelines-distributors elements
    	xPath = ConfigHelper.DISTRIBUTOR_ELEMENT;
    	elements = dc.getElements(xPath);
    	
    	// for each element, create and then load a DistributorConfig instance
    	Iterator<Element> distIt = elements.iterator();
    	PipelinesDistributorConfig config;
    	Element dist;
    	int i = 1;
    	while (distIt.hasNext()) {
    		// get the distributor element
    		dist = distIt.next();
    		
    		// create a new distributor config
    		config = new PipelinesDistributorConfig();
    		
    		// and load it
    		config.load(dist, i, dc);
    		config.setInstLogInt(this.getInstrumentorLogInterval());
    		config.setMasterConfig(this);
    		
    		// finally, add the loaded config to the list
    		distributors.add(config);
    		
    		// increment the distributor index counter
    		i++;
    	}
    	
    }

	private void setupRelays() throws ConfigException {
		PipelinesDistributorConfig relayDist;
		PipelinesConnectorConfig relayConn;
		
		// for each distributor
		for (PipelinesDistributorConfig distConfig : this.getDistributors()) {
			
			// check each service for any relay pipelines
			for (PipelineServiceConfig srvConfig : distConfig.getServices()) {
				
				for (PipelineConfig plConfig : srvConfig.getPipelines()) {
		
					// lookup the given connector on the given distributor
					if (plConfig.getType().equals(PipelineConfig.RELAY_PIPELINE_TYPE)) {
						
						// get the relay to distributor reference; if null, we have an invalid send-to value
						relayDist = ConfigHelper.getDistributorConfigByName(this, plConfig.getSendTo());
						if (relayDist == null) {
							// TODO: add valid distributors to reference at end of message?
							throw new ConfigException("Invalid send-to distributor reference '" + plConfig.getSendTo() + 
									"' on pipeline " + plConfig.getName() + " for service " + srvConfig.getName() + " in distributor " +
									distConfig.getName());
						}
						
						// then get the connector on the given relay distributor; if null, we have an invalid connector reference
						relayConn = ConfigHelper.getConnectorConfigByName(relayDist, plConfig.getConnector());
						if (relayConn == null) {
							throw new ConfigException ("No " + plConfig.getConnector() + " connector configured on send-to ref'd distributor " +
									relayDist.getName());
						}
						plConfig.setConnConfig(relayConn);
					}
					
				}
			}
			
		}
		
	}
    
    public File getConfigFile() {
		return configFile;
	}

	public void setConfigFile(File configFile) {
		this.configFile = configFile;
	}

	public List<PipelinesDistributorConfig> getDistributors() {
		return distributors;
	}

	public void setDistributors(List<PipelinesDistributorConfig> distribs) {
		this.distributors = distribs;
	}
	
	public boolean equals(Object obj) {
		// check that we have a PipelinesConfig instance
		if (!(obj instanceof PipelinesMasterConfig)) {
			return false;
		}
		PipelinesMasterConfig config = (PipelinesMasterConfig) obj;
		
		// for now, dont check the config file, as the contents of the file are really what we are validating
		//if (this.getConfigFile().getPath() != config.getConfigFile().getPath()) {
		//	return false;
		//}
		
		// and finally, check the distributors lists
		if (this.getDistributors().size() != config.getDistributors().size()) {
			return false;
		}
		for (PipelinesDistributorConfig thisDistConfig : this.getDistributors()) {
			if (!config.getDistributors().contains(thisDistConfig)) {
				return false;
			}
		}
		
		return true;
	}

	public int getInstrumentorLogInterval() {
		return instrumentorLogInterval;
	}

	public void setInstrumentorLogInterval(int instrumentorLogInterval) {
		this.instrumentorLogInterval = instrumentorLogInterval;
	}
}
