package org.softwarepipelines.config.connector;

import java.net.MalformedURLException;
import java.net.URL;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.util.ConfigHelper;
import org.w3c.dom.Element;

public class PipelinesConnectorConfig {
	final private static Logger LOG = Logger.getLogger( PipelinesConnectorConfig.class );
	
	// Connector Type Valid Values
	final public static String LOCAL_CONNECTOR_TYPE = "local";
	final public static String SOCKET_CONNECTOR_TYPE = "socket";
	final public static String WS_CONNECTOR_TYPE = "ws";
	
	/* Distributor Config */
	protected PipelinesDistributorConfig distConfig;
	
	/* Connector Name; for references from services/clients that connect to the parent Distributor */
	protected String name;
	
	/* Type; must be either local, socket or ws */
	protected String type;
	
	/* Host and port to send messages over */
	protected String sendToHost;
	protected int sendToPort;
	
	/* Host and port through which we will receive responses (for client side connectors) */
	protected String responseHost;
	protected int responsePort;
	
	/* URL for web service connectors */
	protected URL url;
	
	public PipelinesConnectorConfig() {
	}
	
	public void load(final Element connector) throws ConfigException {
		// load the name
		this.setName(ConfigHelper.loadReqAttribute(connector, ConfigHelper.NAME_ATTR));
		
		// and then the type
		this.setType(ConfigHelper.loadReqAttribute(connector, ConfigHelper.TYPE_ATTR));
		
		// load and validate the config based on the type specified
		if (type.equals(LOCAL_CONNECTOR_TYPE)) {
			loadLocalConn(connector);
		} else if (type.equals(SOCKET_CONNECTOR_TYPE)) {
			loadSocketConn(connector);
		} else if (type.equals(WS_CONNECTOR_TYPE)) {
			loadWebServiceConn(connector);
		} else {
			// Log an error and throw an exception
			LOG.error("Invalid Pipelines Connector type '" + type + "' on connector '" + name + "'");
			throw new ConfigException("Invalid Pipelines Connector type '" + type + "' on connector '" + name + "'");
		}
	}
	
	private void loadLocalConn(final Element conn) throws ConfigException {
		// nothing more necessary for now; this is a place-holder method in case further config is needed later
	}
	
	private void loadSocketConn(final Element conn) throws ConfigException {
		// load the destination socket host
		this.setSendToHost(ConfigHelper.loadReqAttribute(conn, ConfigHelper.SEND_TO_HOST_ATTR));
		
		// and then the desination socket port
		this.setSendToPort(Integer.parseInt(ConfigHelper.loadReqAttribute(conn, ConfigHelper.SEND_TO_PORT_ATTR)));
		
		// TODO: check if we have any return services configured on this distributor; if so, we need to specify
		// the response host and port
		
		// and then the response host and port
		this.setResponseHost(ConfigHelper.loadOptAttribute(conn, ConfigHelper.RESP_HOST_ATTR));
		if (conn.hasAttribute(ConfigHelper.RESP_PORT_ATTR)) {
			this.setResponsePort(Integer.parseInt(conn.getAttribute(ConfigHelper.RESP_PORT_ATTR)));
		} else {
			this.setResponsePort(-1);
		}
	}
	
	private void loadWebServiceConn(final Element conn) throws ConfigException {
		// load the web service URL
		try {
			String url = ConfigHelper.loadReqAttribute(conn, ConfigHelper.URL_ATTR);
			this.setUrl(new URL(url));
		} catch (MalformedURLException mue) {
			LOG.error("Malformed URL specified in config: " + url);
			throw new ConfigException("Malformed URL specified in config: " + url, mue);
		}
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getType() {
		return type;
	}

	public void setType(String type) {
		this.type = type;
	}

	public String getSendToHost() {
		return sendToHost;
	}

	public void setSendToHost(String sendToHost) {
		this.sendToHost = sendToHost;
	}

	public int getSendToPort() {
		return sendToPort;
	}

	public void setSendToPort(int sendToPort) {
		this.sendToPort = sendToPort;
	}

	public String getResponseHost() {
		return responseHost;
	}

	public void setResponseHost(String responseHost) {
		this.responseHost = responseHost;
	}

	public int getResponsePort() {
		return responsePort;
	}

	public void setResponsePort(int responsePort) {
		this.responsePort = responsePort;
	}

	public URL getUrl() {
		return url;
	}

	public void setUrl(URL url) {
		this.url = url;
	}
	
	public PipelinesDistributorConfig getDistConfig() {
		return distConfig;
	}

	public void setDistConfig(PipelinesDistributorConfig distConfig) {
		this.distConfig = distConfig;
	}

	// XXX don't check the distributor configs, or we will deadlock; plus, not really in the spirit of equals here
	public boolean equals(Object obj) {
		// check that the object is the proper type
		if (!(obj instanceof PipelinesConnectorConfig)) {
			return false;
		}
		PipelinesConnectorConfig config = (PipelinesConnectorConfig) obj;
		
		// check the name
		if (!ConfigHelper.checkStringMemsEqual(this.getName(), config.getName())) {
			return false;
		}
		
		// check the type
		if (!ConfigHelper.checkStringMemsEqual(this.getType(), config.getType())) {
			return false;
		}
		
		// check the destination socket host
		if (!ConfigHelper.checkStringMemsEqual(this.getSendToHost(), config.getSendToHost())) {
			return false;
		}
		
		// check the destination socket port
		if (this.getSendToPort() != config.getSendToPort()) {
			return false;
		}
		
		// check the destination socket host
		if (!ConfigHelper.checkStringMemsEqual(this.getResponseHost(), config.getResponseHost())) {
			return false;
		}
		
		// check the destination socket port
		if (this.getResponsePort() != config.getResponsePort()) {
			return false;
		}
		
		// and finally check the URL path
		if (this.getUrl() == null || config.getUrl() == null) {
			if ((this.getUrl() == null) != (config.getUrl() == null)) {
				return false;
			}
		} else {
			if (!ConfigHelper.checkStringMemsEqual(this.getUrl().getPath(), config.getUrl().getPath())) {
				return false;
			}
		}
			
		return true;
	}
}
