package org.softwarepipelines.config.distributor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.DynamicConfig;
import org.softwarepipelines.config.PipelinesMasterConfig;
import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.softwarepipelines.config.util.ConfigHelper;
import org.w3c.dom.Element;

/**
 * DistributorConfig
 * 
 * Distributors configure a list of PipelinesConnectorConfigs and PipelineServiceConfigs.  The DistributorConfig class has
 * convenience methods for adding a service or connector to the appropriate list, and for loading the configuration from a DOM Element.
 * 
 * @author cholm
 */
public class PipelinesDistributorConfig {
	final private static Logger LOG = Logger.getLogger( PipelinesDistributorConfig.class );
	
	final public static String PUSH_PATTERN = "push";
	final public static String PULL_PATTERN = "pull";
	
	final public static String DISTRIBUTOR_XPATH_PREFIX = ConfigHelper.DISTRIBUTOR_ELEMENT + "[";
	final private static String XPATH_CONNECTORS_SUFFIX = "]//" + ConfigHelper.CONNECTOR_ELEMENT;
	final private static String XPATH_SERVICES_SUFFIX = "]//" + ConfigHelper.PIPELINE_SERVICE_ELEMENT;
	
	/* Master Config */
	protected PipelinesMasterConfig masterConfig;
	
	/* Name */
	protected String name;
	
	/* Communication Pattern */
	protected String pattern;
	
	/* Instrumentor Log Interval */
	protected int instLogInt;
	
	/* Connectors */
	protected List<PipelinesConnectorConfig> connectors;
	
	/* Services */
	protected List<PipelineServiceConfig> services;
	
	/* Service Name to Message Class Map */
	protected HashMap<String, Class<?>> serviceToMsgClassMap;
	
	/* Service names list */
	private List<String> serviceNames;
	/* Connector names list */
	private List<String> connectorNames;
    
    public PipelinesDistributorConfig() {
    }
    
    public void load(final Element dist, final int index, final DynamicConfig dc) throws ConfigException {
    	List<Element> elements;
    	Iterator<Element> elIt;
    	Element e;
    	PipelinesConnectorConfig connConfig;
    	PipelineServiceConfig serviceConfig;
    	
    	// instantiate the lists and map
    	connectors = new ArrayList<PipelinesConnectorConfig>();
    	connectorNames = new ArrayList<String>();
    	services = new ArrayList<PipelineServiceConfig>();
    	serviceNames = new ArrayList<String>();
    	serviceToMsgClassMap = new HashMap<String, Class<?>>();
    	
    	// get the distributor name attribute
    	this.setName(ConfigHelper.loadReqAttribute(dist, ConfigHelper.NAME_ATTR));
    	
    	// and then the pattern attribute
    	this.setPattern(ConfigHelper.loadReqAttribute(dist, ConfigHelper.PATTERN_ATTR));
    	
    	// get all of the pipelines-connector elements
    	elements = dc.getElements(DISTRIBUTOR_XPATH_PREFIX + index + XPATH_CONNECTORS_SUFFIX);
    	
    	// for each connector element, create and load a new PipelinesConnectorConfig
    	elIt = elements.iterator();
    	while (elIt.hasNext()) {
    		// get the connector element
    		e = elIt.next();
    		
    		// instantiate the PipelinesConnectorConfig
    		connConfig = new PipelinesConnectorConfig();
    		
    		// and load it
    		connConfig.load(e);
    		
    		// set the distributor config ref
    		connConfig.setDistConfig(this);
    		
    		// finally, add it to the list
    		addConnector(connConfig);
    	}
    	
    	// get all of the pipelines-service elements
    	elements = dc.getElements(DISTRIBUTOR_XPATH_PREFIX + index + XPATH_SERVICES_SUFFIX);
    	
    	// for each service element, create and load a new PipelineServiceConfig
    	elIt = elements.iterator();
    	int i = 1;
    	while (elIt.hasNext()) {
    		// get the connector element
    		e = elIt.next();
    		
    		// instantiate the PipelineServiceConfig
    		serviceConfig = new PipelineServiceConfig();
    		
    		// and set the distributor name
    		serviceConfig.setDistributorName(this.getName());
    		
    		// and load it
    		serviceConfig.load(e, index, i, dc);
    		
    		// add it to the list
    		addService(serviceConfig);
    		
    		// and finally, add the service to message type map entry
    		this.getServiceToMsgClassMap().put(serviceConfig.getName(), serviceConfig.getMessage().getType());
    		
    		i++;
    	}
    }
    
    protected void addConnector(final PipelinesConnectorConfig config) {
    	// validate the services list and config to insert
    	if (connectors == null) {
    		LOG.warn("Attempted to add connector config for connector " + config.getName() + " to a null connectors list; instantiating list...");
    		connectors = new ArrayList<PipelinesConnectorConfig>();
    	} else if (connectorNames == null) {
    		LOG.warn("Attempted to add connector config for connector " + config.getName() + " to a null connector name list; instantiating list...");
    		connectorNames = new ArrayList<String>();
    	} else if (connectorNames.contains(config.getName())) {
    		LOG.warn("Distributor already contains a configuration for the connector named " + config.getName() + 
    				", did not add config: " + config);
    		return;
    	}
    	
    	// add the config to the services list and the name to the service names list
    	connectors.add(config);
    	connectorNames.add(config.getName());
    }
    
    protected void addService(final PipelineServiceConfig config) {
    	// validate the services list and config to insert
    	if (services == null) {
    		LOG.error("Attempted to add service config for service " + config.getName() + " to a null services list.");
    		return;
    	} else if (serviceNames == null) {
    		LOG.warn("Attempted to add service config for service " + config.getName() + " to a null service name list; instantiating list...");
    		serviceNames = new ArrayList<String>();
    	} else if (serviceNames.contains(config.getName())) {
    		LOG.warn("Distributor already contains a configuration for the service named " + config.getName() + 
    				", did not add config: " + config);
    		return;
    	}
    	
    	// add the config to the services list and the name to the service names list
    	services.add(config);
    	serviceNames.add(config.getName());
    }

	public List<PipelinesConnectorConfig> getConnectors() {
		return connectors;
	}

	public void setConnectors(List<PipelinesConnectorConfig> connectors) {
		this.connectors = connectors;
	}

	public List<PipelineServiceConfig> getServices() {
		return services;
	}

	public void setServices(List<PipelineServiceConfig> services) {
		this.services = services;
	}

	public List<String> getServiceNames() {
		return serviceNames;
	}

	public void setServiceNames(List<String> serviceNames) {
		this.serviceNames = serviceNames;
	}

	public List<String> getConnectorNames() {
		return connectorNames;
	}

	public void setConnectorNames(List<String> connectorNames) {
		this.connectorNames = connectorNames;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getPattern() {
		return pattern;
	}

	public void setPattern(String pattern) {
		this.pattern = pattern;
	}
	
	public int getInstLogInt() {
		return instLogInt;
	}

	public void setInstLogInt(int instLogInt) {
		this.instLogInt = instLogInt;
	}

	public PipelinesMasterConfig getMasterConfig() {
		return masterConfig;
	}

	public void setMasterConfig(PipelinesMasterConfig masterConfig) {
		this.masterConfig = masterConfig;
	}

	public boolean equals (Object obj) {
		// check that the object is an instance of PipelinesDistributorConfig as well
		if (!(obj instanceof PipelinesDistributorConfig)) {
			return false;
		}
		PipelinesDistributorConfig config = (PipelinesDistributorConfig) obj;
		
		// check the name and pattern
		if (!ConfigHelper.checkStringMemsEqual(this.getName(), config.getName())) {
			return false;
		}
		if (!ConfigHelper.checkStringMemsEqual(this.getPattern(), config.getPattern())) {
			return false;
		}
		
		// and then check the connectors
		for (PipelinesConnectorConfig thisConnConfig : this.getConnectors()) {
			if (!config.getConnectors().contains(thisConnConfig)) {
				return false;
			}
		}
		
		// finally, check the services (no need to check the neames, as this is part of the services check
		for (PipelineServiceConfig thisServConfig : this.getServices()) {
			if (!config.getServices().contains(thisServConfig)) {
				return false;
			}
		}
		
		return true;
	}

	public HashMap<String, Class<?>> getServiceToMsgClassMap() {
		return serviceToMsgClassMap;
	}

	public void setServiceToMsgClassMap(
			HashMap<String, Class<?>> serviceToMsgClassMap) {
		this.serviceToMsgClassMap = serviceToMsgClassMap;
	}
}
