package org.softwarepipelines.config.pipeline;

import java.util.List;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.DynamicConfig;
import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.softwarepipelines.config.util.ConfigHelper;
import org.w3c.dom.Element;

public class PipelineConfig {
	final private static Logger LOG = Logger.getLogger( PipelineConfig.class );

	final public static String XPATH_INDEXED_PIPELINE_PREFIX =  "]//" + ConfigHelper.PIPELINE_ELEMENT + "[";
	final public static String LOCAL_PIPELINE_TYPE = "local";
	final public static String RELAY_PIPELINE_TYPE = "relay";
	
	final public static int DEF_INQ_CAP = 1000;
	final public static int DEF_EXEC_TASK_Q_CAP = 100;
	
	/* Service Config */
	protected PipelineServiceConfig serviceConfig;
	
	/* Name */
	protected String name;
	
	/* Type */
	protected String type;
	
	/* Send To */
	protected String sendTo;
	
	/* Connector */
	protected String connector;
	
	/* Connector */
	protected PipelinesConnectorConfig connConfig;
	
	/* Match - For String Expr content based pipelines */
	protected Pattern match;
	
	/* Match From and Match To - For Int Expr content based pipelines */
	protected int matchFrom;
	protected int matchTo;
	
	public PipelineConfig() {
	}
	
	public void load(final Element pipeline, final int distIndex, 
			final int serviceIndex, final int pipelineIndex, final DynamicConfig dc) throws ConfigException {
		
		// check for int-expr or string-expr
		checkExpr(distIndex, serviceIndex, pipelineIndex, dc);	
		
		// load the name
		this.setName(ConfigHelper.loadReqAttribute(pipeline, ConfigHelper.NAME_ATTR));
		
		// load the type
		this.setType(ConfigHelper.loadReqAttribute(pipeline, ConfigHelper.TYPE_ATTR));
		
		// load the appropriate pipeline type
		if (type.equals(LOCAL_PIPELINE_TYPE)) {
			loadLocal(pipeline);
		} else if (type.equals(RELAY_PIPELINE_TYPE)) {
			loadRelay(pipeline);
		} else {
			throw new ConfigException("Invalid service pipeline type: " + type);
		}
		
	}

	private void checkExpr(final int distIndex, final int serviceIndex, final int pipelineIndex, final DynamicConfig dc) throws ConfigException {
		List<Element> els1, els2;
		Element e;
		
		// check for the string expression
		els1 = dc.getElements(PipelinesDistributorConfig.DISTRIBUTOR_XPATH_PREFIX + distIndex + 
    			PipelineServiceConfig.SERVICE_XPATH_INDEX_PREFIX + serviceIndex + XPATH_INDEXED_PIPELINE_PREFIX + pipelineIndex + "]//" +
    			ConfigHelper.STRING_EXPR_ELEMENT);
		if (els1.size() == 0) {
			
			// if we have no elements, check for an int-expr element
			els1 = dc.getElements(PipelinesDistributorConfig.DISTRIBUTOR_XPATH_PREFIX + distIndex + 
	    			PipelineServiceConfig.SERVICE_XPATH_INDEX_PREFIX + serviceIndex + XPATH_INDEXED_PIPELINE_PREFIX + pipelineIndex + "]//" +
	    			ConfigHelper.INT_EXPR_ELEMENT);
			
			// if we have no int-expr element, then we must not be associated with a content based router
			if (els1.size() == 0) {
				//throw new ConfigException("Service pipeline '" + this.getName() + "' is associated with a content based router " +
				//		" and thus must define either a string-expr OR int-expr child element!");
				return;
			}
			// else if we have more than one, throw
			else if (els1.size() > 1) {
				throw new ConfigException("No more than one int-expr element allowed per service pipeline '" + this.getName() + "'!");
			}
			// otherwise, we have a single int-expr; set the match from and match to attrs
			else {
				e = els1.get(0);
				this.setMatchFrom(Integer.parseInt(ConfigHelper.loadReqAttribute(e, ConfigHelper.MATCH_FROM_ATTR)));
				this.setMatchTo(Integer.parseInt(ConfigHelper.loadReqAttribute(e, ConfigHelper.MATCH_TO_ATTR)));
			}
			
		} else if (els1.size() == 1) {
			
			// else if we have exactly one element, make sure we don't have an int-expr element
			els2 = dc.getElements(PipelinesDistributorConfig.DISTRIBUTOR_XPATH_PREFIX + distIndex + 
	    			PipelineServiceConfig.SERVICE_XPATH_INDEX_PREFIX + serviceIndex + XPATH_INDEXED_PIPELINE_PREFIX + pipelineIndex + "]//" +
	    			ConfigHelper.INT_EXPR_ELEMENT);
			if (els2.size() != 0) {
				throw new ConfigException("No int-expr elements allowed once a string-expr has been defined on pipeline '" + 
						this.getName() + "'!");
			}
			
			// and then set the match Pattern
			e = els1.get(0);
			this.setMatch(Pattern.compile(ConfigHelper.loadReqAttribute(e, ConfigHelper.MATCH_ATTR)));
			
		} else {
			
			// otherwise we have more than one string-expr element, which is invalid (for now)
			throw new ConfigException("No more than one string-expr element allowed per service pipeline '" + this.getName() + "'!");
		}
	}
	
	private void loadLocal(final Element pipeline) throws ConfigException {
		// if we have a connector or sendTo set, log that it will be ignored
		if (pipeline.hasAttribute(ConfigHelper.SEND_TO_ATTR)) {
			LOG.info("send-to value ignored for local pipeline " + this.getName());
		}
		if (pipeline.hasAttribute(ConfigHelper.CONNECTOR_ATTR)) {
			LOG.info("connector value ignored for local pipeline " + this.getName());
		}
	}
	
	private void loadRelay(final Element pipeline) throws ConfigException {
		// load the required sendTo attribute (references a distributor in the same config file)
		this.setSendTo(ConfigHelper.loadReqAttribute(pipeline, ConfigHelper.SEND_TO_ATTR));
		
		// and then load the required connector attribute
		this.setConnector(ConfigHelper.loadReqAttribute(pipeline, ConfigHelper.CONNECTOR_ATTR));
	}
	
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public PipelineServiceConfig getServiceConfig() {
		return serviceConfig;
	}

	public void setServiceConfig(PipelineServiceConfig serviceConfig) {
		this.serviceConfig = serviceConfig;
	}

	public String getType() {
		return type;
	}

	public void setType(String type) {
		this.type = type;
	}

	public String getSendTo() {
		return sendTo;
	}

	public void setSendTo(String sendTo) {
		this.sendTo = sendTo;
	}

	public String getConnector() {
		return connector;
	}

	public void setConnector(String connector) {
		this.connector = connector;
	}

	public Pattern getMatch() {
		return match;
	}

	public void setMatch(Pattern match) {
		this.match = match;
	}

	public int getMatchFrom() {
		return matchFrom;
	}

	public void setMatchFrom(int matchFrom) {
		this.matchFrom = matchFrom;
	}

	public int getMatchTo() {
		return matchTo;
	}

	public void setMatchTo(int matchTo) {
		this.matchTo = matchTo;
	}

	public PipelinesConnectorConfig getConnConfig() {
		return connConfig;
	}

	public void setConnConfig(PipelinesConnectorConfig connConfig) {
		this.connConfig = connConfig;
	}

	public boolean equals(Object obj) {
		// TODO: Log equals() fail points
		
		if (!(obj instanceof PipelineConfig)) {
			return false;
		}
		PipelineConfig config = (PipelineConfig) obj;
		
		// check that the routers are equal
		//if (!this.getRouter().equals(config.getRouter())) {
		//	return false;
		//}
		
		// check the name
		if (!ConfigHelper.checkStringMemsEqual(this.getName(), config.getName())) {
			return false;
		}
		
		// check the type
		if (!ConfigHelper.checkStringMemsEqual(this.getType(), config.getType())) {
			return false;
		}
		
		// check the send to
		if (!ConfigHelper.checkStringMemsEqual(this.getSendTo(), config.getSendTo())) {
			return false;
		}
		
		// check the connector
		if (!ConfigHelper.checkStringMemsEqual(this.getConnector(), config.getConnector())) {
			return false;
		}
		
		// check the match
		if (this.getMatch() == null || config.getMatch() == null) {
			if ((this.getMatch() == null) != (config.getMatch() == null)) {
				return false;
			}
		} else {
			if (!ConfigHelper.checkStringMemsEqual(this.getMatch().pattern(), config.getMatch().pattern())) {
				return false;
			}
		}
		
		// check the match from
		if (this.getMatchFrom() != config.getMatchFrom()) {
			return false;
		}
		
		// and finally check the match to
		if (this.getMatchTo() != config.getMatchTo()) {
			return false;
		}
		
		return true;
	}
}
