package org.softwarepipelines.config.service;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.softwarepipelines.config.DynamicConfig;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.message.MessageConfig;
import org.softwarepipelines.config.router.PipelinesRouterConfig;
import org.softwarepipelines.config.util.ConfigHelper;
import org.softwarepipelines.config.pipeline.PipelineConfig;
import org.w3c.dom.Element;

/**
 * Configuration information for a PipelineService.
 *
 *
 * @author cholm
 *
 */
public class PipelineServiceConfig {
	
	final public static String SERVICE_XPATH_INDEX_PREFIX = "]/" + ConfigHelper.PIPELINE_SERVICE_ELEMENT + "[";
	final private static String XPATH_SERVICE_MSG_SUFFIX = "]//" + ConfigHelper.MESSAGE_ELEMENT;
	final private static String XPATH_SERVICE_ROUTER_SUFFIX = "]//" + ConfigHelper.ROUTER_ELEMENT;
	final private static String XPATH_SERVICE_PIPELINES_SUFFIX = "]//" + ConfigHelper.PIPELINE_ELEMENT;
	
    /* Service Name */
    protected String name;
    
    /* Service Name */
    protected String className;
    
    /* Distributor Name */
    protected String distributorName;
    
    /* Sequential Message Processing Required? */
    protected boolean oneWay;
    
    /* Message */
    protected MessageConfig message;
    
    /* Router */
    protected PipelinesRouterConfig router;
    
    /* Pipelines */
    protected List<PipelineConfig> pipelines;

    /**
     * All PipelineServices must have service name constructor-injected.
     * 
     * Other dependencies will be setter-injected.
     * 
     * @param aName
     */
    public PipelineServiceConfig() {
    }
    
    public void load(final Element service, final int parentIndex, final int index, final DynamicConfig dc) throws ConfigException {
    	MessageConfig msgConfig;
    	
    	// load the service name
    	this.setName(ConfigHelper.loadReqAttribute(service, ConfigHelper.NAME_ATTR));
    	
    	// load the service class name
    	this.setClassName(ConfigHelper.loadReqAttribute(service, ConfigHelper.CLASS_NAME_ATTR));
    	
    	// and then whether the service is one way (true) or request-response (false)
    	this.setOneWay(Boolean.parseBoolean(ConfigHelper.loadReqAttribute(service, ConfigHelper.ONE_WAY_ATTR)));
    	
    	// instantiate and load a new Message config
    	msgConfig = loadMessage(parentIndex, index, dc);
    	this.setMessage(msgConfig);
    	
    	// instantiate and load a new Router config
    	loadRouter(msgConfig, parentIndex, index, dc);
    	
    	// get all of the service pipelines config elements
    	loadPipelines(parentIndex, index, dc);
    	
    }
    
    private void loadRouter(final MessageConfig msgConfig, final int parentIndex, final int index, final DynamicConfig dc) throws ConfigException {
    	PipelinesRouterConfig rtrConfig = new PipelinesRouterConfig();
    	rtrConfig.setMessageConfig(msgConfig);
    	rtrConfig.setServiceConfig(this);
    	
    	List<Element> elements = dc.getElements(PipelinesDistributorConfig.DISTRIBUTOR_XPATH_PREFIX + parentIndex + 
    			SERVICE_XPATH_INDEX_PREFIX + index + XPATH_SERVICE_ROUTER_SUFFIX);
    	if (elements.size() != 1) {
    		throw new ConfigException("Must configure exactly one router per pipeline service!  Configured: " + elements.size());
    	}
    	rtrConfig.load(elements.get(0));
    	this.setRouter(rtrConfig);
    }
    
    private MessageConfig loadMessage(final int parentIndex, final int index, final DynamicConfig dc) throws ConfigException {
    	MessageConfig msgConfig = new MessageConfig();
    	
    	List<Element> elements = dc.getElements(PipelinesDistributorConfig.DISTRIBUTOR_XPATH_PREFIX + parentIndex + 
    			SERVICE_XPATH_INDEX_PREFIX + index + XPATH_SERVICE_MSG_SUFFIX);
    	if (elements.size() != 1) {
    		throw new ConfigException("Must configure exactly one message per pipeline service!  Configured: " + elements.size());
    	}
    	msgConfig.load(elements.get(0));
    	return msgConfig;
    }

    private void loadPipelines(final int parentIndex, final int index, final DynamicConfig dc) throws ConfigException {
    	Element e;
    	Iterator<Element> elIt;
    	List<Element> elements;
    	PipelineConfig plConfig;
    	
    	// instantiate the pipelines list
    	this.pipelines = new ArrayList<PipelineConfig>();
    	
    	// get the pipelines elements
    	elements = dc.getElements(PipelinesDistributorConfig.DISTRIBUTOR_XPATH_PREFIX + parentIndex + 
    			SERVICE_XPATH_INDEX_PREFIX + index + XPATH_SERVICE_PIPELINES_SUFFIX);
    	
    	// and for each instance, instante and load the configs
    	elIt = elements.iterator();
    	int i = 1;
    	while (elIt.hasNext()) {
    		// get the pipeline config
    		e = elIt.next();
    		
    		// create a new instance of the pipeline config
    		plConfig = new PipelineConfig();
    		
    		// and set the service class name
    		plConfig.setServiceConfig(this);
    		
    		// and load it
    		plConfig.load(e, parentIndex, index, i, dc);
    		
    		// TODO: add convenience method (similar to addService and addConnector in dist config) to check for dups, etc
    		this.pipelines.add(plConfig);
    		i++;
    	}
    	this.setPipelines(this.pipelines);
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }
    
    public String getClassName() {
		return className;
	}

	public void setClassName(String className) {
		this.className = className;
	}
	
	public boolean isOneWay() {
		return oneWay;
	}

	public void setOneWay(boolean oneWay) {
		this.oneWay = oneWay;
	}

	public MessageConfig getMessage() {
		return message;
	}

	public void setMessage(MessageConfig message) {
		this.message = message;
	}

	public PipelinesRouterConfig getRouter() {
		return router;
	}

	public void setRouter(PipelinesRouterConfig router) {
		this.router = router;
	}

	public List<PipelineConfig> getPipelines() {
		return pipelines;
	}

	public void setPipelines(List<PipelineConfig> pipelines) {
		this.pipelines = pipelines;
	}
	
	public String getDistributorName() {
		return distributorName;
	}

	public void setDistributorName(String distributorName) {
		this.distributorName = distributorName;
	}

	public boolean equals(Object obj) {
        if (!(obj instanceof PipelineServiceConfig)) {
            return false;
        }
        PipelineServiceConfig config = (PipelineServiceConfig) obj;
        
        // check the name
        if (!ConfigHelper.checkStringMemsEqual(this.getName(), config.getName())) {
			return false;
		}
        
        // check the distributor name
        if (!ConfigHelper.checkStringMemsEqual(this.getDistributorName(), config.getDistributorName())) {
			return false;
		}
        
        // check if one way
        if (this.isOneWay() != config.isOneWay()) {
        	return false;
        }
        
        // check the message
        if (!this.getMessage().equals(config.getMessage())) {
        	return false;
        }
        
        // check the router
        if (!this.getRouter().equals(config.getRouter())) {
        	return false;
        }
        
        // and finally check the pipelines
        for (PipelineConfig plConfig : this.getPipelines()) {
        	if (!config.getPipelines().contains(plConfig)) {
        		return false;
        	}
        }
        
        return true;
    }

    public String toString() {
        return 
        	"[PipelineServiceConfig]" + "\n\tserviceName:\t" + this.getName() +
        	"\n\tclassName:\t" + this.getClassName() +
        	"\n\\toneWay?:\t" + this.isOneWay() +
        	"\n\tmessage:\t" + this.getMessage().toString() +
        	"\n\trouter:\t" + this.getRouter().toString() +
        	"\n\tpipelines:\t" + this.getPipelines();
    }
}
