package org.softwarepipelines.config.util;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.PipelinesMasterConfig;
import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.pipeline.PipelineConfig;
import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.w3c.dom.Element;

public class ConfigHelper {
	final private static Logger LOG = Logger.getLogger(ConfigHelper.class);
	
	/* Config Constants */
	// Common attrs
	public static final String CLASS_NAME_ATTR = "class-name";
	public static final String ID_ATTR = "id";
	public static final String SEND_TO_HOST_ATTR = "send-to-host";;
	public static final String SEND_TO_PORT_ATTR = "send-to-port";
	public static final String RESP_HOST_ATTR = "response-host";
	public static final String RESP_PORT_ATTR = "response-port";
	public static final String NAME_ATTR = "name";
	public static final String PATTERN_ATTR = "pattern";
	public static final String TYPE_ATTR = "type";
	public static final String URL_ATTR = "url";
	
	// Root element
	public static final String PIPELINES_CONFIG_ELEMENT = "pipelines-config";
	
	// Instrumentor
	public static final String INSTRUMENTOR_ELEMENT = "instrumentor";
	public static final String LOG_INT_ATTR = "log-interval";
	
	// Distributors
    public static final String DISTRIBUTOR_ELEMENT = "pipelines-distributor";
    
    // Connectors
	public static final String CONNECTOR_ELEMENT = "pipelines-connector";
    
    // Pipeline Services
    public static final String PIPELINE_SERVICE_ELEMENT = "pipeline-service";
    public static final String ONE_WAY_ATTR = "one-way";

    // Pipelines Routers
	public static final String ROUTER_ELEMENT = "pipelines-router";
	public static final String ENFORCE_SEQUENCE_ATTR = "enforce-sequence";
	public static final String ROUTE_METHOD_ATTR = "route-method";
	
	// Messages
	public static final String MESSAGE_ELEMENT = "message";
	public static final String PIPELINE_KEY_ATTR = "pipeline-key";
	
	// Pipelines
	public static final String PIPELINE_ELEMENT = "pipeline";
	public static final String SEND_TO_ATTR = "send-to";
	public static final String CONNECTOR_ATTR = "connector";
	public static final String MIN_THREADS_ATTR = "min-threads";
	public static final String MAX_THREADS_ATTR = "max-threads";
	public static final String STRING_EXPR_ELEMENT = "string-expr";
	public static final String MATCH_ATTR = "match";
	public static final String INT_EXPR_ELEMENT = "int-expr";
	public static final String MATCH_FROM_ATTR = "match-from";
	public static final String MATCH_TO_ATTR = "match-to";
	
	// Exception messages
	final public static String NO_ROUTER_MSG = "No parent router set!";
	
	public static PipelinesDistributorConfig getDistributorConfigByName(final PipelinesMasterConfig config, final String distName) {
		for (PipelinesDistributorConfig distConfig : config.getDistributors()) {
			if (distConfig.getName().equals(distName)) {
				return distConfig;
			}
		}
		return null;
	}
	
	public static PipelinesConnectorConfig getConnectorConfigByName(final PipelinesDistributorConfig distConfig, final String name) throws ConfigException {
		for (PipelinesConnectorConfig connConfig : distConfig.getConnectors()) {
			if (connConfig.getName().equals(name)) {
				
				// check to make sure that the connector referenced is not a local type
				//if (connConfig.getType().equals(PipelinesConnectorConfig.LOCAL_CONNECTOR_TYPE)) {
				//	throw new ConfigException("Relay pipelines can not use local connectors to relay messages!  Create additional " +
				//			"pipelines in this distributor, or reference another connector in the relay destination distributor"); 
				//}
				return connConfig;
			}
		}
		
		return null;
	}
	
	/**
	 * Returns the socket connector configured in the distributor config, or null if none is specified.
	 */
	public static PipelinesConnectorConfig getConnectorConfigByType(final PipelinesDistributorConfig distConfig, final String connType) {
		PipelinesConnectorConfig toRet = null;
		
		for (PipelinesConnectorConfig config : distConfig.getConnectors()) {
			if (config.getType().equals(connType)) {
				toRet = config;
				break;
			}
		}
		return toRet;
	} 
	
	public static boolean isDistributorServiceOneWay(final PipelinesDistributorConfig distConfig, final String serviceName) throws ConfigException {
		for (PipelineServiceConfig srvConfig : distConfig.getServices()) {
			if (srvConfig.getName().equals(serviceName)) {
				return srvConfig.isOneWay();
			}
		}
		throw new ConfigException("No service named " + serviceName + " configured on distributor " + distConfig.getName());
	}
	
	public static boolean isDistributorServiceRelay(final PipelineServiceConfig config) throws ConfigException {
		// check each pipeline for the relay type
		for (PipelineConfig plConfig : config.getPipelines()) {
			if (plConfig.getType().equals(PipelineConfig.RELAY_PIPELINE_TYPE)) {
				return true;
			}
		}
		return false;
	}
	
	public static PipelineConfig getRelayPipeline(final PipelineServiceConfig config) throws ConfigException {
		for (PipelineConfig plConfig: config.getPipelines()) {
			if (plConfig.getType().equals(PipelineConfig.RELAY_PIPELINE_TYPE)) {
				return plConfig;
			}
		}
		LOG.warn("No relay type pipeline found!");
		return null;
	}

	public static String loadReqAttribute(final Element element, final String attrName) throws ConfigException {
		if (element.hasAttribute(attrName)) {
			return element.getAttribute(attrName); 
		} else {
			throw new ConfigException("No attribute '" + attrName + "' set on element: " + element.getNodeName());
		}
	}
	
	public static String loadOptAttribute(final Element element, final String attrName) throws ConfigException {
		if (element.hasAttribute(attrName)) {
			return element.getAttribute(attrName); 
		} else {
			return null;
		}
	}
	
	public static boolean checkStringMemsEqual(final String s1, final String s2) {
		if (s1 == null) {
			return s2 == null;
		} else {
			return s1.equals(s2);
		}
	}
	
}
