package org.softwarepipelines.config;

import junit.framework.TestCase;

import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.message.MessageConfig;
import org.softwarepipelines.config.router.PipelinesRouterConfig;
import org.softwarepipelines.config.pipeline.PipelineConfig;
import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.softwarepipelines.config.util.ConfigHelper;

import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;


public class DistributorConfigTestCase extends TestCase {
	
	private PipelinesMasterConfig golden, config;
	
    public DistributorConfigTestCase(String name) throws FileNotFoundException, IOException, ConfigException {
        super(name);
        
        // load the golden pipelines config from the application context
        ApplicationContext ac = new ClassPathXmlApplicationContext("config/org/softwarepipelines/framework/test/goldenConfig.xml");
        
        // load the two configurations
        golden = (PipelinesMasterConfig) ac.getBean("pipelines-config");
        config = new PipelinesMasterConfig(new File(
        	"config/org/softwarepipelines/framework/test/two-layer-pipelines-config.xml"));
        
        // check the == operator is false to ensure not the same object
    	assertFalse(golden == config);
    }
    
    /**
     * Tests the DistributorProperties loadProps method.
     */
    public void testLoadConfigDistributors() {
    	
        // validate the distributor list
    	assertEquals("Configs should have the same number of distributors!", golden.getDistributors().size(), config.getDistributors().size());
    	
    	PipelinesDistributorConfig dist;
    	int i = 0;
    	for (PipelinesDistributorConfig goldenDist : golden.getDistributors()) {
    		// get the framework configured distributor
    		dist = config.getDistributors().get(i);
    		
	    	// validate the distributors
    		validateDistributors(goldenDist, dist);
	    	
	    	// increment the counter
	    	i++;
    	}
    }
    
    /**
     * Tests the DistributorProperties loadProps method.
     */
    public void testLoadConfigConnectors() {
    	
    	// get the distributors
    	List<PipelinesDistributorConfig> goldDists = golden.getDistributors();
    	List<PipelinesDistributorConfig> confDists = config.getDistributors();
    	
    	// for each distributor list, get the connectors and validate them against one another
    	PipelinesDistributorConfig dist;
    	PipelinesConnectorConfig conn;
    	int i = 0, j = 0;
    	for (PipelinesDistributorConfig golden : goldDists) {
    		
    		// get the configured distributor
    		dist = confDists.get(i);
    		
    		// and check that we have the correct number of connectors
    		assertEquals("Distributors should have the same number of connectors!", golden.getConnectors().size(), dist.getConnectors().size());
    		
    		// get all connectors for each of the distributors, and validate them
    		for (PipelinesConnectorConfig goldenConn : golden.getConnectors()) {
    			conn = dist.getConnectors().get(j);
    			
    			validateConnectors(goldenConn, conn);
    			
    			j++;
    		}
    		j = 0;
    		i++;
    	}
    }
    
    /**
     * Tests the DistributorProperties loadProps method.
     */
    public void testLoadConfigServices() {
    	
    	// get the distributors
    	List<PipelinesDistributorConfig> goldDists = golden.getDistributors();
    	List<PipelinesDistributorConfig> confDists = config.getDistributors();
    	
    	// for each distributor list, get the connectors and validate them against one another
    	PipelinesDistributorConfig dist;
    	PipelineServiceConfig serv;
    	int i = 0, j = 0;
    	for (PipelinesDistributorConfig golden : goldDists) {
    		
    		// get the configured distributor
    		dist = confDists.get(i);
    		
    		// and check that we have the correct number of connectors
    		assertEquals("Distributors should have the same number of services!", golden.getServices().size(), dist.getServices().size());
    		
    		// get all connectors for each of the distributors, and validate them
    		for (PipelineServiceConfig goldenServ : golden.getServices()) {
    			serv = dist.getServices().get(j);
    			
    			validateServices(goldenServ, serv);
    			
    			j++;
    		}
    		j = 0;
    		i++;
    	}
    }

    /**
     * Tests the DistributorProperties loadProps method.
     */
    public void testConfigEqualsMethods() throws FileNotFoundException, IOException, ConfigException {
    	// fine-grained equals checks, in case of failures; check top level .equals at the end, as if it fails, the error message will be useless
    	// distributor checks
    	
    		// connector checks
    	
    		// service checks
    	
    			// router check
    			
    			// message check
    			
    			// pipelines checks
    	
    	// finally, check the pipelines config top level equals
    	assertTrue(golden.equals(config));
    }
    
    private void validateDistributors(final PipelinesDistributorConfig goldenDist, final PipelinesDistributorConfig dist) {
    	// check the names
    	assertTrue("Distributor names should be equal: " + goldenDist.getName() + "," + dist.getName(), 
    			ConfigHelper.checkStringMemsEqual(goldenDist.getName(), dist.getName()));
    	
    	// check the patterns
    	assertTrue("Distributor patterns should be equal: " + goldenDist.getPattern() + "," + dist.getPattern(), 
    			ConfigHelper.checkStringMemsEqual(goldenDist.getPattern(), dist.getPattern()));
    	
    	// finally, check the service and connector name lists
    	assertEquals("Distributor connector name lists should be the same!", goldenDist.getConnectorNames(), dist.getConnectorNames());
    	assertEquals("Distributor service name lists should be the same!", goldenDist.getServiceNames(), dist.getServiceNames());
    }
    
    private void validateConnectors(final PipelinesConnectorConfig goldenConn, final PipelinesConnectorConfig conn) {
    	// check the names
    	assertTrue("Connector names should be equal: " + goldenConn.getName() + "," + conn.getName(), 
    			ConfigHelper.checkStringMemsEqual(goldenConn.getName(), conn.getName()));
    	
    	// check the types
    	assertTrue("Connector types should be equal: " + goldenConn.getType() + "," + conn.getType(), 
    			ConfigHelper.checkStringMemsEqual(goldenConn.getType(), conn.getType()));
    	
    	// check the send to hosts
    	assertTrue("Connector hosts should be equal: " + goldenConn.getSendToHost() + "," + conn.getSendToHost(), 
    			ConfigHelper.checkStringMemsEqual(goldenConn.getSendToHost(), conn.getSendToHost()));
    	
    	// check the send to ports
    	assertEquals("Connector ports should be equal!", goldenConn.getSendToPort(), conn.getSendToPort());
    	
    	// check the response hosts
    	assertTrue("Connector hosts should be equal: " + goldenConn.getResponseHost() + "," + conn.getResponseHost(), 
    			ConfigHelper.checkStringMemsEqual(goldenConn.getSendToHost(), conn.getSendToHost()));
    	
    	// check the response ports
    	assertEquals("Connector ports should be equal!", goldenConn.getResponsePort(), conn.getResponsePort());
    	
    	// and finally check the urls
    	assertEquals("Connector URLs should be equal!", goldenConn.getUrl(), conn.getUrl());
    }
    
    private void validateServices(final PipelineServiceConfig goldenServ, final PipelineServiceConfig serv) {
    	// check the names
    	assertTrue("Service names should be equal: " + goldenServ.getName() + "," + serv.getName(), 
    			ConfigHelper.checkStringMemsEqual(goldenServ.getName(), serv.getName()));
    	
    	// and the one way values
    	assertEquals("Services should have the same one way setting!", goldenServ.isOneWay(), serv.isOneWay());
    	
    	// validate the service routers
    	validateRouters(goldenServ.getRouter(), serv.getRouter());
    	
    	// validate the service messages
    	validateMessages(goldenServ.getMessage(), serv.getMessage());
    	
    	// and finally, validate the service pipelines
    	validatePipelines(goldenServ.getPipelines(), serv.getPipelines());
    }
    
    private void validateRouters(final PipelinesRouterConfig goldRtr, final PipelinesRouterConfig confRtr) {
    	// check the route-method
    	//assertTrue("Route route-methods should be equal: " + goldRtr.getRouteMethod() + "," + confRtr.getRouteMethod(), 
    	//		ConfigHelper.checkStringMemsEqual(goldRtr.getRouteMethod(), confRtr.getRouteMethod()));
    	
    	// check the class-name
    	assertTrue("Route class-names should be equal: " + goldRtr.getClassName() + "," + confRtr.getClassName(), 
    			ConfigHelper.checkStringMemsEqual(goldRtr.getClassName(), confRtr.getClassName()));
    	
    	// and whether or not both enforce sequence
    	assertEquals("Routers should have the same enforce-sequence setting!", goldRtr.isEnforceSequence(), confRtr.isEnforceSequence());
    }
    
    private void validateMessages(final MessageConfig goldMsg, final MessageConfig confMsg) {
    	// validate the parent routers
    	//validateRouters(goldMsg.getRouter(), confMsg.getRouter());
    	
    	// check the type
    	assertTrue("Message types should be equal: " + goldMsg.getType() + "," + confMsg.getType(), 
    			ConfigHelper.checkStringMemsEqual(goldMsg.getType().getName(), confMsg.getType().getName()));
    	
    	// and the pipeline key
    	assertTrue("Message pipeline-keys should be equal: " + goldMsg.getPipelineKey() + "," + confMsg.getPipelineKey(), 
    			ConfigHelper.checkStringMemsEqual(goldMsg.getPipelineKey(), confMsg.getPipelineKey()));
    }
    
    private void validatePipelines(final List<PipelineConfig> goldPls, final List<PipelineConfig> confPls) {
    	// check that we have the same number
    	assertEquals("Service pipeline lists should be the same size!", goldPls.size(), confPls.size());
    	
    	// then check that the configured pipelines list contains each of the golden pipelines
    	for (PipelineConfig goldPl : goldPls) {
    		// TODO: impl toString in the service pipeline config class
    		assertTrue("Configured service pipeline list does not contain golden pipeline: " + goldPl.getName(), confPls.contains(goldPl));
    	}
    }
}
