package org.softwarepipelines.examples.hello.ch20;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.softwarepipelines.examples.hello.HelloMessage;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.service.PipelineServiceBase;

/**
 * The HelloPipelinesDatabaseService uses a technique for partitioning 
 * database calls over a number of DBMS databases, servers or instances. 
 * It utilizes the dbShards JDBC driver to perform sharding transparently 
 * to the application. The service invokes the normal JDBC methods to
 * INSERT rows into the database, and the driver "shards" the statements
 * accordingly, based on the shard-config.xml file.
 * 
 * @author Andy Grove, cholm
 * @created Jun 17, 2008
 */
public class HelloPipelinesDatabaseService extends PipelineServiceBase {
	final private static Logger LOG = Logger.getLogger(HelloPipelinesDatabaseService.class);
	Connection conn = null;
	final String jdbcUrl = "jdbc:dbshards:source/examples/src/org/softwarepipelines/examples/hello/ch21/conf/";
	
	/**
	 * Required constructor which creates the driver instance and 
	 * opens a connection to the sharding JDBC driver. Note that
	 * the actual physical database connections are managed by
	 * the sharding JDBC driver.
	 * @param config
	 * @throws PipelineException
	 */
	public HelloPipelinesDatabaseService(PipelineServiceConfig config) throws PipelineException {
		super(config);
		
		/* Initialize the database sharding driver.*/
		try {
			Class.forName( "com.dbshards.jdbc.Driver" ).newInstance();
		
			/* Open a connection using the driver.*/
			conn = DriverManager.getConnection(jdbcUrl, "user", "password");
			conn.setAutoCommit(false);
		} catch (Exception e) {
			throw new PipelineException("Error initializing the database sharding connection!");
		}
		
	}
	
	/** 
	 * Receive the HelloMessage, parse the first and last name
	 * values, and insert the person into the database. 
	 */
	/* (non-Javadoc)
	 * @see org.softwarepipelines.framework.service.PipelineServiceBase#processMessage(org.softwarepipelines.framework.message.AbstractMessage)
	 */
	@Override
	public void processMessage(AbstractMessage msg) throws PipelineException {
		/* Cast the incoming message to a HelloMessage */
		if (!(msg instanceof HelloMessage)) {
			throw new PipelineException("Service " + this.getName() + " requires " + HelloMessage.class.getName());
		}
		HelloMessage helloMsg = (HelloMessage) msg;
		
		/* Parse the first and last names from the full name in the message */
		String [] names = helloMsg.getName().split(" ");
		try {
			insertPerson(names[0], names[1]);
		} catch (Exception e) {
			throw new PipelineException("Error inserting person: " + helloMsg.getName());
		}
		
		/* Send a confirmation message */
		helloMsg.setName("Software Pipelines Database");
		sendResultMessage(helloMsg);
	}

	/**
	 * Insert the person into the database using the sharding JDBC driver.
	 * @param firstName
	 * @param lastName
	 * @throws Exception
	 */
	private void insertPerson(String firstName, String lastName) throws Exception {

		PreparedStatement stmt = conn
				.prepareStatement("INSERT INTO person (first_name, last_name) VALUES (?, ?)");
		stmt.setString(1, firstName);
		stmt.setString(2, lastName);
		int rows = stmt.executeUpdate();

		LOG.debug("updated: " + rows + " row(s)");

		stmt.close();

		conn.commit();
	}
	
	//public void finalize() {
	public void shutdown() {
		
		try {
			/* Close the database sharding connection.*/
			conn.close();
		} catch (Exception e) {
			// No Exception needed from finalize method.
		}
	}
}
