package org.softwarepipelines.examples.pbcor;

import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.softwarepipelines.examples.pbcor.AccountTransMessage;
import org.softwarepipelines.examples.pbcor.CheckBalance;
import org.softwarepipelines.examples.pbcor.DebitAcctMessage;
import org.softwarepipelines.framework.connector.PipelinesConnector;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.service.PipelineServiceBase;
import org.softwarepipelines.framework.util.PipelinesHelper;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

public class ValidateTransactionService extends PipelineServiceBase {
	
	/* Connector through which we will invoke the debit account service */
	protected PipelinesConnector connector;
	
	/* Target Service; the next service in the invoke chain */
	protected String targetService;
	
	/* The check balance service we will invoke */
	protected CheckBalance checkBalance;

	public ValidateTransactionService(PipelineServiceConfig config) throws PipelineException {
		super(config);
		
		/* Get the Spring application context */
		ApplicationContext ac = new ClassPathXmlApplicationContext("pbcorAppContext.xml");
		
		/* Set the target service */
		targetService = "debitAcctService";
		
		/* Load the connector from the destination map configured in the application context. */
		connector = PipelinesHelper.getConnectorFromDestMap(ac, getDestMapKey(), targetService);
		
		/* Instantiate the check balance service */
		checkBalance = new CheckBalance();
	}

	public void processMessage(AbstractMessage msg) throws PipelineException {
		AccountTransMessage acctMsg;
		
		/* Validate that we have an AccountTrans message (or the appropriate message type) */
		if (!(msg instanceof AccountTransMessage)) {
			throw new PipelineException("ValidateTransaction service requires an " + AccountTransMessage.class.getName() + " message!");
		}
		acctMsg = (AccountTransMessage) msg;
		
		/* Check the transaction type (either by type or by checking the string) and validate accordingly */
		if (acctMsg instanceof DebitAcctMessage) {
			
			acctMsg.setValid(validateDebitTrans((DebitAcctMessage)acctMsg));
			
		} else {
			throw new PipelineException(ValidateTransactionService.class.getName() + " service only supports " + 
					DebitAcctMessage.class.getName() + " messages!");
		}
		
		/* If the message is valid, invoke the debit account service */
		if (acctMsg.isValid()) {
			acctMsg.setServiceName(targetService);
			connector.send(acctMsg);
		}  
		/* Otherwise, send the invalid transaction message (with explanation message) back as the result message */
		else {
			sendResultMessage(acctMsg);
		}
	}
	
	/**
	 * Validate the debit account transaction.
	 * 
	 * @param msg the debit account message
	 * @return whether or not the transaction is valid
	 * @throws PipelineException
	 */
	private boolean validateDebitTrans(DebitAcctMessage msg) throws PipelineException {
		
		/* Call check balance */
		float balance = checkBalance.getBalance(msg.getAcctNum());
		
		/* If the balance can not cover the debit amount, attach the explanation message and return false */
		if (balance < msg.getAmount()) {
			msg.setMsg("Insufficient funds!  Current account balance is $" + balance);
			return false;
		} 
		/* Otherwise, return true */
		else {
			return true;
		}
	}
	
	/**
	 * Finally, implement our shutdown logic
	 */
	public void shutdown() throws PipelineException {
		super.shutdown();
		connector.shutdown();
		checkBalance.shutdown();
	}
}
