package org.softwarepipelines.examples.pbcor.acctdb;

import java.sql.SQLException;
import org.softwarepipelines.framework.exception.PipelineException;
import javax.sql.DataSource;
import java.sql.ResultSet;
import org.springframework.jdbc.core.simple.SimpleJdbcTemplate;
import org.springframework.jdbc.core.simple.ParameterizedRowMapper;
import org.springframework.transaction.annotation.Transactional;

/**
 * Implement of the accountDB using Spring JDBC.
 */
@Transactional
public class AccountDBJDBC implements ParameterizedRowMapper<Account>, AccountDB {
	
	/* Spring JDBC template object.*/
	protected SimpleJdbcTemplate jdbcTemplate;

	/* Data source for connection to the database.*/
	protected DataSource dataSource;

	/**
	 * Default constructor.
	 */
	public AccountDBJDBC() {
	}

	/**
	 * This method is invoked by the Spring JDBC framework to inject the datasource,
	 * and initialze the JDBC template.
	 * 
	 * @param dataSource
	 */
	public void setDataSource(DataSource dataSource)
	{
		this.dataSource = dataSource;
		jdbcTemplate = new SimpleJdbcTemplate(dataSource);
	}

	/**
	 * Map result rows to an Account object.
	 *
	 * @param rs
	 * @param row
	 * @throws SQLException
	 * @return Account
	 */
	public Account mapRow(ResultSet rs, int row) throws SQLException
	{
		Account dto = new Account();
		dto.setAcctNum( rs.getInt( 1 ) );
		dto.setBranchId( rs.getInt( 2 ) );
		dto.setPin( rs.getInt( 3 ) );
		dto.setBalance( rs.getFloat( 4 ) );
		return dto;
	}

	/**
	 * Retrieve an account by number.
	 */
	public Account getAccountByNumber(long acctNum) throws PipelineException {
		
		// check if the account map contains our key; if so, return the account
		/**
		 * Returns all rows from the account table that match the criteria 'branch_id = :branchId AND account_no = :accountNo'.
		 */
		try {
			Account account = jdbcTemplate.queryForObject("SELECT account_no, branch_id, pin, balance FROM account WHERE account_no = ?", this, acctNum);
			return account;
		}
		catch (Exception e) {
			throw new PipelineException("Query failed", e);
		}
	}
	
	/**
	 * Save an account.
	 */
	public void saveAccount(Account dto) throws PipelineException {
		try {	
			jdbcTemplate.update("UPDATE account SET account_no = ?, branch_id = ?, pin = ?, balance = ? WHERE account_no = ?", dto.getAcctNum(), dto.getBranchId(), dto.getPin(), dto.getBalance(), dto.getAcctNum());
		} catch (Exception e) {
			throw new PipelineException("Update failed", e);
		}
	}
	
	/**
	 * Shutdown method.
	 */
	public void shutdown() {
		/* Empty implementation.*/
	}
}
