package org.softwarepipelines.framework.callback;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.concurrent.ArrayBlockingQueue;

import org.apache.log4j.Logger;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.nio.OpsChangeRequest;
import org.softwarepipelines.framework.util.MessageHelper;

// TODO: refactor to connect to the connector awaiting our response; after connection established, prepare the
// buffer to be written through the Broker
public class PipelinesSocketCallbackHandlerImpl implements PipelinesCallbackHandler {
	final private static Logger LOG = Logger.getLogger(PipelinesSocketCallbackHandlerImpl.class);
	
	// maximum connect retries
	public static final int CONNECT_TIMEOUT = 30000;
	public static final int MAX_CONN_TRIES = 50;
	
	/* Selector Ref */
	protected Selector selector;
	
	/* SocketChannel */
	protected SocketChannel channel;
	
	/* Message Buffer Reference */
	protected ByteBuffer msgBuff;
	
	/* Pending Changes Queue */
	protected ArrayBlockingQueue<OpsChangeRequest> pendingChanges;
    
    /* Destination Host */
    protected String host;
    
    /* Destination Port */
    protected int port;

	public PipelinesSocketCallbackHandlerImpl() {
	}
	
	/**
	 * Opens and connects a SocketChannel, and registers it for write operations. 
	 * 
	 * @param aHost the host to which we will try to connect the socket channel
	 * @param aPort the port on which we will attempt to connect
	 */
	public void connectAndReg(final String aHost, final int aPort, final Selector aSelector, final ArrayBlockingQueue<OpsChangeRequest> changeQ) throws PipelineException {
		// set the host and port
		this.host = aHost;
		this.port = aPort;
		
		// open the SocketChannel
		try {
			channel = SocketChannel.open();
			channel.configureBlocking(false);
		} catch (IOException ioe) {
			throw new PipelineException("Error opening channel for socket callback handler on host:port (" + host + ":" + port + ")");
		}
		
		// and try to connect to the destination broker
    	int tries = 0;
    	while (!channel.isConnected() && tries < MAX_CONN_TRIES) {
        	try {
	            if (channel.isConnectionPending()) {
	                channel.finishConnect();
	            } else if (!channel.isConnected()) {
	                channel.connect(new InetSocketAddress(host, port));
	            }
        	} catch (IOException ioe) {
        	}
        	try {Thread.sleep(CONNECT_TIMEOUT/MAX_CONN_TRIES);} catch (Exception e) {}
        	tries++;
        }
    	
    	// see if we timed out
    	if (tries >= MAX_CONN_TRIES) {
    		throw new PipelineException("Timed out establishing connection to host:port (" + host + ":" + port + ")");
    	}
        
        // finally, queue a change to register for write ops with the selector
    	msgBuff = ByteBuffer.allocate(2048);
    	//msgBuff.flip();
		OpsChangeRequest change = new OpsChangeRequest();
		change.setType(OpsChangeRequest.REGISTER_CHANGE_TYPE);
		change.setAttachment(msgBuff);
		change.setChannel(channel);
		//change.setOps(SelectionKey.OP_WRITE);
		change.setOps(0);
		
		// set the pending changes queue
		pendingChanges = changeQ;
		while (!pendingChanges.contains(change)) {
			try {
				pendingChanges.put(change);
				break;
			} catch (InterruptedException ie) {
				LOG.warn("Interrupted when putting change request on change queue!");
			}
		}
		
		selector = aSelector;
		selector.wakeup();
	}

	// TODO: change to check the message for the sender information
	public void returnResult(AbstractMessage result) throws PipelineException {
    	LOG.debug("[" + Thread.currentThread().getName() + "] receiveResult in Connector " + this.toString());
        
        // check that we have a message buffer
        if (getMsgBuff() == null) {
            throw new PipelineException("Null currKey SelectionKey reference when attempting to forward response!");
        }

        // and stuff the response message into it
        msgBuff.clear();
        MessageHelper.writeMessageToBuffer(result, msgBuff);
        
        // and declare for write ops
        OpsChangeRequest change = new OpsChangeRequest();
        change.setType(OpsChangeRequest.OP_CHANGE_TYPE);
        change.setChannel(channel);
        change.setAttachment(msgBuff);
        change.setOps(SelectionKey.OP_WRITE);
        while (!pendingChanges.contains(change)) {
			try {
				pendingChanges.put(change);
				break;
			} catch (InterruptedException ie) {
				LOG.warn("Interrupted when putting change request on change queue!");
			}
		}
        
        selector.wakeup();
        

        // finally, get the key's SC and register it for write ops
        //SocketChannel sc = (SocketChannel) getCurrKey().channel();

        //try {
        //    SelectionKey newKey = sc.register(getCurrKey().selector(), SelectionKey.OP_WRITE);
        //    newKey.attach(keyBuf);
        //    getCurrKey().selector().wakeup();
            
            // TODO: set the current key to the new key?
            
        //} catch (ClosedChannelException cce) {
        //    LOG.warn("Channel closed when trying to register to write response!",
        //        cce);
        //}
    }
	 
	public String getHost() {
		return host;
	}

	public void setHost(String host) {
		this.host = host;
	}

	public int getPort() {
		return port;
	}

	public void setPort(int port) {
		this.port = port;
	}

	public ByteBuffer getMsgBuff() {
		return msgBuff;
	}

	public void setMsgBuff(ByteBuffer msgBuff) {
		this.msgBuff = msgBuff;
	}
}
