package org.softwarepipelines.framework.connector;

import org.apache.log4j.Logger;

import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.util.ConfigHelper;
import org.softwarepipelines.framework.callback.PipelinesCallbackHandler;
import org.softwarepipelines.framework.distributor.PipelinesDistributorBase;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;


/**
 * PipelineServiceInvoker that invokes services at the local API level.
 *
 * @author cholm
 *
 */
public class LocalPipelinesConnector extends PipelinesConnectorBase implements PipelinesCallbackHandler {
    private static final Logger LOG = Logger.getLogger(LocalPipelinesConnector.class);

    /* Distributor Instance */
    protected PipelinesDistributorBase distributor;

    public LocalPipelinesConnector(PipelinesDistributorBase dist)
        throws PipelineException {
        super();

        distributor = dist;
    }
    
    public LocalPipelinesConnector(final PipelinesConnectorConfig config) throws PipelineException {
    	super(config);
    }

    public void send(AbstractMessage msg) throws PipelineException {
    	LOG.debug("Sending message: " + msg);
    	
    	// TODO: throw exception if we are invoking a return service as a void service?
    	
        // invoke the service in the distributor
        distributor.distribute(msg);
    }

    public AbstractMessage sendAndReceive(final AbstractMessage msg) throws PipelineException {
    	LOG.debug("Sending and receiving response for message: " + msg);
    	
    	// check to ensure that we are invoking a valid return service; if not, throw
    	try {
	    	if (ConfigHelper.isDistributorServiceOneWay(distributor.getConfig(), msg.getServiceName())) {
	    		throw new PipelineException(PipelineException.NOT_A_RETURN_SERVICE_MSG +
	    				"\n\tDistributor: " + distributor.getConfig().getName() +
	    				"\n\tService: " + msg.getServiceName());
	    	}
    	} catch (ConfigException ce) {
    		throw new PipelineException("Error checking if service " + msg.getServiceName() + " on distributor " + 
    				distributor.getConfig().getName() + " is one way", ce);
    	}
    	
        // set the connector reference and invoke the service in the distributor
    	msg.setConnectorRef(this.getConnectorRef().toString());
        distributor.distribute(msg);

        // and wait for the response on the incoming message queue
        return this.waitForResponse();
    }
    
    /**
     * Shut down this connector.
     * 
     * Check the connector back into the distributor (cleans up map entries and such).
     */
    public void shutdown() {
    	((PipelinesDistributorBase)distributor).checkinLocalConnector(this);
    }
    
    /**
     * Enqueue the incoming message on our incoming msgs queue
     */
    public void returnResult(final AbstractMessage result) throws PipelineException {
        // place the incoming result AbstractMessage onto the incomingMsgs queue to wake up the invokeReturnService thread
        do {
            try {
                this.getIncomingMsgs().put(result);
                LOG.debug("Put result message on incoming message queue: " + result);
                break;
            } catch (InterruptedException ie) {
                LOG.warn("Interrupted while enqueueing received result: " +
                    result, ie);
            }
        } while (!this.getIncomingMsgs().contains(result));
    }

    public PipelinesDistributorBase getDistributor() {
        return distributor;
    }

	public void setDistributor(PipelinesDistributorBase distributor) {
		this.distributor = distributor;
	}
}
