package org.softwarepipelines.framework.connector;

import org.apache.log4j.Logger;

import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.nio.ClientMessageBroker;
import org.softwarepipelines.framework.nio.OpsChangeRequest;
import org.softwarepipelines.framework.util.MessageHelper;

import java.io.IOException;

import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.concurrent.TimeUnit;

// TODO: refactor to use a ServerSocketChannel to accept connections for multi tiered responses
// should function more like the MessageBroker, or perhaps we need a ClientMessageBroker as well?
public class NIOPipelinesConnector extends PipelinesConnectorBase {
    private final static Logger LOG = Logger.getLogger(NIOPipelinesConnector.class);

    /* Maximum Response Retires */
    public static final int RESP_TIMEOUT = 5000;
    public static final int MAX_RESP_TRIES = 50;
    public static final int MAX_CONN_TRIES = 50;
    
    /* Client Message Broker Selector */
    protected ClientMessageBroker broker;

    /* SocketChannels */
    protected SocketChannel outChannel;
    
    /* Write Buffers */
    protected ByteBuffer outMsgBuff;
    
    /**
     * 
     * @param config
     * @throws PipelineException
     */
    public NIOPipelinesConnector(final PipelinesConnectorConfig config) throws PipelineException, ClosedChannelException {
    	// call the super class constructor
        super(config);

        // and then open the outgoing socket channel 
        try {
            outChannel = SocketChannel.open();
            outChannel.configureBlocking(false);
        } catch (IOException ioe) {
            LOG.error("IOException opening and configuring SocketChannel", ioe);
            throw new PipelineException("IOException opening and configuring SocketChannel",
                ioe);
        }
        
        // finally, allocate the message buffers
        outMsgBuff = ByteBuffer.allocate(2048);
    }

    @Override
    public void send(AbstractMessage msg) throws PipelineException {
        // make sure we are connected to the broker
    	if (!outChannel.isConnected()) {
    		this.connect();
    	}
    	
    	// XXX
    	// TODO: flesh this out. this may cause problems as currently we are assuming that no other connector that touches this message
    	// will do so by calling sendAndReceive (that is, any relay or other service invoking connectors in the middle of
    	// the pipelining infrastructure should wait for a response to the message it sends)
    	// XXX
    	// set the connector ref, sender host and sender port
    	//msg.setConnectorRef(getConnectorRef().toString());
    	//msg.setSenderHost(config.getResponseHost());
    	//msg.setSenderPort(config.getResponsePort());
    	
    	// TODO: wait for the message buffer to be cleared (after it is written over the socket?)
    	// or flip it in the client broker?
    	
    	// Write the message into the buffer
    	int pos = outMsgBuff.position();
    	while (pos == outMsgBuff.position()) {
    		MessageHelper.writeMessageToBuffer(msg, outMsgBuff);
    	}
    	
    	// if the pending changes list is already full, wake up the 
    	// selector to deal with them before adding our own
    	//getBroker().getSelector().wakeup();
    	
    	// and then declare an interest in write ops if we have no pending change for this channel
    	OpsChangeRequest change = new OpsChangeRequest();
    	change.setAttachment(outMsgBuff);
    	change.setChannel(outChannel);
    	change.setType(OpsChangeRequest.OP_CHANGE_TYPE);
    	change.setOps(SelectionKey.OP_WRITE);
    	if (!getBroker().getPendingChanges().contains(change)) {
    		getBroker().enqueueChangeRequest(change);
    	}
    	
    	// and wake up the selecting thread
    	getBroker().getSelector().wakeup();
    }

    @Override
    public AbstractMessage sendAndReceive(AbstractMessage msg) throws PipelineException {
    	//LOG.debug("[" + Thread.currentThread().getName() + "] invokeReturn in Connector " + this.toString());
    	// first, set the connector reference
    	//msg.setConnectorRef(this.getConnectorRef().toString());
    	
    	// XXX
    	// TODO: flesh this out. this may cause problems as currently we are assuming that no other connector that touches this message
    	// will do so by calling sendAndReceive (that is, any relay or other service invoking connectors in the middle of
    	// the pipelining infrastructure should wait for a response to the message it sends)
    	// XXX
    	// set the connector ref, sender host and sender port
    	msg.setConnectorRef(getConnectorRef().toString());
    	msg.setSenderHost(config.getResponseHost());
    	msg.setSenderPort(config.getResponsePort());
    	
        // send the message
    	send(msg);
    	
    	// wait for the response on the incoming message queue
    	AbstractMessage result = null;
    	while (result == null) {
    		try {
    			result = getIncomingMsgs().poll(RESP_TIMEOUT, TimeUnit.MILLISECONDS);
    			if (result == null) {
    				LOG.error("Timed out waiting for response to message: " + msg);
    				throw new PipelineException(PipelineException.RESP_TIMEOUT_MSG + "\tSent message: " + msg);
    			} else {
    				break;
    			}
    		} catch (InterruptedException ie) {
    			LOG.warn("Interrupted waiting for response, retrying...");
    		}
    	}
    	
    	return result;
    }

    /**
     * Connect the SocketChannel to the PipelinesDistributor.
     */
    public void connect() throws PipelineException {
    	// connect to the server message broker
    	while (!outChannel.isConnected()) {
	    	try {
		    	if (outChannel.isConnectionPending()) {
		        	outChannel.finishConnect();
		        } else if (!outChannel.isConnected()) {
		            outChannel.connect(new InetSocketAddress(config.getSendToHost(), config.getSendToPort()));
		        }
	    	} catch (IOException ioe) {
	    		throw new PipelineException("Error connecting socket channel!", ioe);
	    	}
    	}
    	
    	// clear the message buffer to attach a fresh buffer
    	outMsgBuff.clear();
    	
    	// and create the pending change, registering for the empty op set initially (when we are ready to write, we
    	// will declare an interest)
    	OpsChangeRequest change = new OpsChangeRequest();
    	change.setChannel(outChannel);
    	change.setType(OpsChangeRequest.REGISTER_CHANGE_TYPE);
    	change.setOps(0);
    	change.setAttachment(outMsgBuff);
    	getBroker().enqueueChangeRequest(change);
    	
    	// and wake up the selector
    	getBroker().getSelector().wakeup();
    }

    public void shutdown() throws PipelineException {
    	// wait for outgoing messages to be written
    	try {Thread.sleep(100);} catch (Exception e) {}
    	
    	// check back into the client message broker
        getBroker().checkinSocketConnector(this);
    	
        // and close the channel
        if (outChannel.isOpen()) {
            try {
                outChannel.close();
            } catch (IOException ioe) {
                throw new PipelineException("Error shutting down connector SocketChannel!",
                    ioe);
            }
        }
    }

	public SocketChannel getOutChannel() {
		return outChannel;
	}

	public void setOutChannel(SocketChannel outChannel) {
		this.outChannel = outChannel;
	}

	public ByteBuffer getOutMsgBuff() {
		return outMsgBuff;
	}

	public void setOutMsgBuff(ByteBuffer outMsgBuff) {
		this.outMsgBuff = outMsgBuff;
	}

	public ClientMessageBroker getBroker() {
		return broker;
	}

	public void setBroker(ClientMessageBroker broker) {
		this.broker = broker;
	}
}
