package org.softwarepipelines.framework.connector;

import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;


/**
 * Interface for a client-side pipelines connector; includes methods for sending a one-way message, sending a message and receiving a response
 * message, and shutting down the connector.<br>
 * <br>
 * Clients use <code>PipelinesConnector</code> instances to send messages to a certain <code>PipelinesDistributor</code>.  Clients should use 
 * the helper method <code>PipelinesHelper.getConnectorByName(String distributorName, String connectorName, PipelinesMasterConfig config)</code> 
 * to obtain a connector with the given name, configured on the distributor with the given name in the pipelines master configuration provided:<br>
 * <br>
 * <code>
 * PipelinesConnector connector = PipelinesHelper.getConnectorByName("myDistributor", "myConnector", masterConfig);
 * </code>
 * 
 * @author cholm
 *
 */
public interface PipelinesConnector {
    /**
     * Invoke a service from which we do not expect a AbstractMessage in return.
     *
     * @param msg the message to send to the service for processing
     */
    public void send(final AbstractMessage msg) throws PipelineException;

    /**
     * Invoke a service from which we expect a AbstractMessage returned.
     *
     * @param msg the message to send the service for processing
     * @return the message reply sent from the invoked service
     */
    public AbstractMessage sendAndReceive(final AbstractMessage msg) throws PipelineException;
    
    
    /**
     * Convenience method to shut down this connector.
     */
    public void shutdown() throws PipelineException;
}
