package org.softwarepipelines.framework.distributor;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.pipeline.PipelineConfig;
import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.softwarepipelines.config.util.ConfigHelper;
import org.softwarepipelines.framework.callback.PipelinesCallbackHandler;
import org.softwarepipelines.framework.connector.LocalPipelinesConnector;
import org.softwarepipelines.framework.connector.PipelinesConnectorFactory;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.nio.ClientMessageBroker;
import org.softwarepipelines.framework.nio.MessageBrokerFactory;
import org.softwarepipelines.framework.nio.ServerMessageBroker;
import org.softwarepipelines.framework.router.PipelinesRouter;
import org.softwarepipelines.framework.router.PipelinesRouterBase;
import org.softwarepipelines.framework.router.PipelinesRouterFactory;
import org.softwarepipelines.framework.util.Instrumentor;

import java.util.HashMap;

/**
 * Base class for PipelinesDistributor instances.  Implements the {@link org.softwarepipelines.framework.distributor.PipelinesDistributor} interface.
 *
 * @author cholm
 *
 */
public abstract class PipelinesDistributorBase implements PipelinesDistributor {
	private static final Logger LOG = Logger.getLogger( PipelinesDistributorBase.class );
	
	/* Client Message Broker (for relaying messages to other distributors) */
	protected ClientMessageBroker clientBroker;
	
	/* Server Message Broker (for incoming messages) */
	protected ServerMessageBroker serverBroker;
	
	/* Distributor Config */
    protected PipelinesDistributorConfig config;
	
    /* Service Name To Router Map */
	protected HashMap<String, PipelinesRouter> serviceToRouterMap;
	
	/* Callback Ref to Callback Map */
	protected HashMap<String, PipelinesCallbackHandler> connectorRefToCallbackMap;
	
	/* Instrumentor Log Interval */
	protected int instLogInterval; 

    public PipelinesDistributorBase() throws PipelineException {
        // instantiate the maps
    	serviceToRouterMap = new HashMap<String, PipelinesRouter>();
    	connectorRefToCallbackMap = new HashMap<String, PipelinesCallbackHandler>();
    }
    
    /**
     * Configure this distributor instance.
     * 
     * PipelinesDistributors will configure one or more PipelinesConnectors and exactly one PipelinesRouter for each service hosted.
     */
    public void configure(final PipelinesDistributorConfig config) throws PipelineException {
    	// set the config
    	this.config = config;
    	
    	// and set the Instrumentor log interval
    	Instrumentor.setLogInterval(config.getInstLogInt());
    	
    	// if have have a socket connector configured, create the message broker to receive messages over that connector
    	// TODO: refactor for multiple socket connectors?
    	PipelinesConnectorConfig socketConnConfig = ConfigHelper.getConnectorConfigByType(
    			config, PipelinesConnectorConfig.SOCKET_CONNECTOR_TYPE);
    	if (socketConnConfig != null) {
    		// TODO: refactor to pass dist reference into factory?
    		setServerBroker(MessageBrokerFactory.getMessageBroker(config));
    		getServerBroker().setDistributor(this);
    	}
    	
    	// and then configure a router for each service
    	this.serviceToRouterMap = new HashMap<String, PipelinesRouter>();
    	PipelinesRouterBase router;
    	PipelineConfig relayConfig;
    	for (PipelineServiceConfig servConfig : config.getServices()) {
    		// if we have not defined a client message broker, and this service has any relay pipelines, create the
    		// client message broker now
    		try {
	    		if (clientBroker == null && ConfigHelper.isDistributorServiceRelay(servConfig)) {
	    			
	    			// get the relay config
	    			relayConfig = ConfigHelper.getRelayPipeline(servConfig);
	    			if (relayConfig == null) {
	    				throw new PipelineException("Helper said this was a relay service, no relay pipeline found!");
	    			}
	    			
	    			// create a new client broker with the connector configuration on the relay pipeline
	    			clientBroker = MessageBrokerFactory.getClientMessageBroker(relayConfig.getConnConfig().getDistConfig());
	    		}
    		} catch (ConfigException ce) {
    			throw new PipelineException("Error checking if service " + servConfig.getName() + " is relay type", ce);
    		}
    		
    		router = PipelinesRouterFactory.getPipelinesRouter(servConfig.getRouter(), this);
    		this.getServiceToRouterMap().put(servConfig.getName(), router);
    	}
    }
    
    /**
     * Start the distributor's client broker and message broker, if they are not presently alive.
     */
    public void start() throws PipelineException {
    	
    	if (getClientBroker() != null && !getClientBroker().isAlive()) {
    		clientBroker.start();
    		LOG.info("Started client message broker on distributor " + this.getConfig().getName());
    	}
    	
    	if (getServerBroker() != null && !getServerBroker().isAlive()) {
    		serverBroker.start();
    		LOG.info("Started server message broker on distributor " + this.getConfig().getName());
    	}
    }

    /**
     * Shut down each pipline's thread pool executor.
     */
    public void shutdown() throws PipelineException {
    	PipelinesRouter rtr;

    	// shut down each service router
        for (String srvName : this.getConfig().getServiceNames()) {
        	rtr = this.getServiceToRouterMap().get(srvName);
        	((PipelinesRouterBase)rtr).shutdown();
        }
        
        // shutdown the brokers
        if (clientBroker != null) {
        	clientBroker.shutdown();
        }
        if (serverBroker != null) {
        	serverBroker.shutdown();
        }
        
        // and then shut ourselves down in the factory
        PipelinesDistributorFactory.shutdownDistributor(this);
        
        // shut down the instrumentor
        Instrumentor.shutdown();
    }
    
	/**
	 * Convenience method for adding a connector UUID ref -> PipelinesCallbackHandler map entry
	 * 
	 * @param ref the connector UUID
	 * @param callback the callback handler
	 * @throws PipelineException if something goes wrong
	 */
	public void addConnectorToCallbackRef(final String ref, final PipelinesCallbackHandler callback) throws PipelineException {
		// check to make sure the map doesn't contain that reference
		if (this.getConnectorRefToCallbackMap().containsKey(ref)) {
			throw new PipelineException("Map already contains UUID connector reference: " + ref);
		}
		
		// and add the entry
		this.getConnectorRefToCallbackMap().put(ref, callback);
	}
	
	/*========================================= 
	 * CONNECTOR CHECKIN AND CHECKOUT METHODS 
	 * ========================================*/
	/**
	 * Used by services hosted in this distributor to checkout connectors to other distributors in the pipelines configuration.
	 *
	public NIOPipelinesConnector checkoutSocketConnector(final String distName) throws PipelineException {
		// look up the distributor in the config
	}*/
	
	public LocalPipelinesConnector checkoutLocalConnector() throws PipelineException {
		LocalPipelinesConnector localConn;
		
		// iterate through the connector configs to see if we have a local connector configured
		for (PipelinesConnectorConfig connConfig : this.getConfig().getConnectors()) {
			if (connConfig.getType().equals(PipelinesConnectorConfig.LOCAL_CONNECTOR_TYPE)) {
				
				// if we find a connector of the specified type, create an instance with the factory
				localConn = (LocalPipelinesConnector) PipelinesConnectorFactory.getPipelinesConnector(connConfig);
				localConn.setDistributor(this);
				
				// add the callback reference to the map
				this.addConnectorToCallbackRef(localConn.getConnectorRef().toString(), localConn);
				
				// and finally, return
				return localConn;
			}
		}
		
		// if we made it here, we do not have a local connector configuration
		throw new PipelineException("No local connector configured on distributor " + this.getConfig().getName());
	}
	
	public void checkinLocalConnector(final LocalPipelinesConnector conn) {
		// check that we have an entry for the UUID ref
		if (!this.getConnectorRefToCallbackMap().containsKey(conn.getConnectorRef().toString())) {
			LOG.warn("No entry in map for connector " + conn.getConfig().getName() + ", ignoring checkin");
		}
		
		// and remove the entry
		this.getConnectorRefToCallbackMap().remove(conn.getConnectorRef().toString());
	}
	
	
	/* GETTERS AND SETTERS */
	public PipelinesDistributorConfig getConfig() {
		return config;
	}

	public void setConfig(PipelinesDistributorConfig config) {
		this.config = config;
	}

	public HashMap<String, PipelinesRouter> getServiceToRouterMap() {
		return serviceToRouterMap;
	}

	public HashMap<String, PipelinesCallbackHandler> getConnectorRefToCallbackMap() {
		return connectorRefToCallbackMap;
	}

	public void setConnectorRefToCallbackMap(
			HashMap<String, PipelinesCallbackHandler> connectorRefToCallbackMap) {
		this.connectorRefToCallbackMap = connectorRefToCallbackMap;
	}

	public void setServiceToRouterMap(
			HashMap<String, PipelinesRouter> serviceToRouterMap) {
		this.serviceToRouterMap = serviceToRouterMap;
	}

	public int getInstLogInterval() {
		return instLogInterval;
	}

	public void setInstLogInterval(int instLogInterval) {
		this.instLogInterval = instLogInterval;
	}

	public ClientMessageBroker getClientBroker() {
		return clientBroker;
	}

	public void setClientBroker(ClientMessageBroker clientBroker) {
		this.clientBroker = clientBroker;
	}

	public ServerMessageBroker getServerBroker() {
		return serverBroker;
	}

	public void setServerBroker(ServerMessageBroker serverBroker) {
		this.serverBroker = serverBroker;
	}
}
