package org.softwarepipelines.framework.distributor;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.framework.exception.PipelineException;

import java.util.HashMap;

/**
 *
 * @author Andy Grove, cholm
 *
 */
public class PipelinesDistributorFactory {
	final private static Logger LOG = Logger.getLogger(PipelinesDistributorFactory.class);
	
	// distributors created in this factory
	private static HashMap<String, PipelinesDistributorBase> distributorMap;
	
	/* Pattern to Class Name Map */
	private static HashMap<String, String> typeToDistClassMap;
	final public static String PUSH_PATTERN_DIST_CLASS_NAME = "org.softwarepipelines.framework.distributor.PushPatternDistributor";
	final public static String PULL_PATTERN_DIST_CLASS_NAME = "org.softwarepipelines.framework.distributor.PullPatternDistributor";
	
	// static init
	static {
		distributorMap = new HashMap<String, PipelinesDistributorBase>();
		
		typeToDistClassMap = new HashMap<String, String>();
		typeToDistClassMap.put(PipelinesDistributorConfig.PUSH_PATTERN, PUSH_PATTERN_DIST_CLASS_NAME);
		typeToDistClassMap.put(PipelinesDistributorConfig.PULL_PATTERN, PULL_PATTERN_DIST_CLASS_NAME);
	}
	
	
	/**
	 * Retrieves (optionally creating) the singleton PipelinesDistributor instance which the config specifies. 
	 * 
	 * @param config the distributor configuration
	 * @return a singleton distributor instance
	 * @throws PipelineException
	 */
    public static PipelinesDistributorBase getPipelinesDistributor(final PipelinesDistributorConfig config) throws PipelineException {
    	PipelinesDistributorBase toRet;
    	
    	// if we have already instantiated this distributor, return a reference
    	if (distributorMap.containsKey(config.getName())) {
    	
    		toRet = distributorMap.get(config.getName());
    		
    	} else {
    	
	    	// otherwise, instantiate the new distributor; first, get the distributor class name based on the pattern
	        String distClassName;
	        if (typeToDistClassMap.containsKey(config.getPattern())) {
	        	distClassName = typeToDistClassMap.get(config.getPattern());
	    	} else {
	    		throw new PipelineException("Distributor pattern " + config.getPattern() + " is unknown and thus not mapped to a distributor class name");
	        }
	
	        try {
	            // create the distributor
	            toRet = (PipelinesDistributorBase) Class.forName(distClassName).newInstance();
	            
	            // and add a map entry
	            distributorMap.put(config.getName(), toRet);
	            
	            // and configure the new distributor
	            toRet.configure(config);
	            
	        } catch (Exception e) {
	            throw new PipelineException("Failed to create the distributor named " + config.getName() + " of type " + distClassName, e);
	        }
    	}
    	return toRet;
    }
    
    /**
     * Shut down the distributor with the given name.
     * 
     * Removes the singleton instance from the static distributor map and calls shutdown().
     * 
     * @param distName the name of the distributor
     */
    public static void shutdownDistributor(final String distName) {
    	if (!distributorMap.containsKey(distName)) {
    		LOG.warn("Distributor '" + distName + 
    				"' is not one of the factory singletons, ignoring factory shutdown request...");
    	}
    	distributorMap.remove(distName);
    }
    
    /**
     * Shut down the distributor with the given name.
     * 
     * Removes the singleton instance from the static distributor map and calls shutdown().
     * 
     * @param distName the name of the distributor
     */
    public static void shutdownDistributor(final PipelinesDistributorBase dist) {
    	shutdownDistributor(dist.getConfig().getName());
    }
}
