package org.softwarepipelines.framework.message;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.Serializable;

import org.softwarepipelines.config.util.ConfigHelper;


/**
 * Base class for all messages sent through the pipelines framework.<br>
 * <br>
 * Custom messages must extend this class, and must set <code>serviceName</code> to the destination service as configured in the 
 * <code>PipelinesMasterConfig</code>.  Custom messages may also specify the <code>opName</code>, to allow for different operations,
 * and thus different behavior, in the destination service.<br>
 * <br>
 * If a custom message is to be sent, at any point, via a socket connector, the custom message class must override the java.io.Externalizable
 * interface methods writeExternal and readExternal, and must call into its parent implementations of these methods first.  See the documentation
 * for the Externalizable methods in this class for further details and examples.<br>
 * <br>
 * 
 * @author cholm, Andy Grove
 */
public abstract class AbstractMessage implements Externalizable, Serializable {
	
	/* data types that have specific type encoding support */
    private static final byte TYPE_SHORT = 1;
    private static final byte TYPE_INT = 2;
    private static final byte TYPE_LONG = 3;
    private static final byte TYPE_FLOAT = 4;
    private static final byte TYPE_DOUBLE = 5;
    private static final byte TYPE_STRING = 6;
    private static final byte TYPE_BYTEARRAY = 7;
    private static final byte TYPE_OBJECT = 8;
    private static final byte TYPE_BOOLEAN = 9;
    private static final byte TYPE_CHAR = 10;
	
	/* Destination Service Name */
    protected String serviceName;
    
    /* Callback Reference */
    private String connectorRef;
    
    /* Sender Host and Port */
    private String senderHost;
    private int senderPort;
    
    /* Operation Name
     * Optional field to specify different operations for a given message 
     **/
    protected String opName;
    
    /**
     * Default constructor; allows for setter injection of message attributes.
     */
    public AbstractMessage() {
    }
    
    /**
     * Constructor that injects the service name at construction.
     * 
     * @param aServName the name of the destination <code>PipelineService</code>
     */
    public AbstractMessage(final String aServName) {
    	this.setServiceName(aServName);
    }
    
    /**
     * Get the name of the destination service.
     * 
     * @return the destination service name
     */
    public String getServiceName() {
		return serviceName;
	}
    
    /**
     * Set the destination service.
     * 
     * @param serviceName the destination service
     */
	public void setServiceName(String serviceName) {
		this.serviceName = serviceName;
	}
	
	/**
	 * Internal use only.
	 * @return
	 */
	public String getSenderHost() {
		return senderHost;
	}

	/**
	 * Internal use only.
	 * @param senderHost
	 */
	public void setSenderHost(String senderHost) {
		this.senderHost = senderHost;
	}

	/**
	 * Internal use only.
	 * @return
	 */
	public int getSenderPort() {
		return senderPort;
	}

	/**
	 * Internal use only.
	 * @param senderPort
	 */
	public void setSenderPort(int senderPort) {
		this.senderPort = senderPort;
	}

	/**
	 * Internal use only.
	 * @return
	 */
	public String getConnectorRef() {
		return connectorRef;
	}

	/**
	 * Internal use only.
	 * @param connectorRef
	 */
	public void setConnectorRef(String connectorRef) {
		this.connectorRef = connectorRef;
	}

	/**
	 * Set the operation name, to be used by the destination service.
	 * @return the operation name
	 */
	public String getOpName() {
		return opName;
	}

	/**
	 * Set the operation name, to be used by the destination service.
	 * @param opName the operation name to set
	 */
	public void setOpName(String opName) {
		this.opName = opName;
	}

	/**
	 * See the java.io.Externalizable interface.
	 * 
	 * An example of how to implement this methods is given below, for a sample custom message class, MyMessage.<br>
	 * <br>
	 * 
	 * <code>
	 * public class MyMessage extends AbstractMessage {<br>
	 * <br>
	 * 	...<br>
	 *	// custom message members and methods, for example:<br>
	 *<br>
	 *	// some integer<br>
	 *	int someInt;<br>
	 *<br>
	 *	// some required string<br>
	 *	String reqdString;<br>
	 *<br>
	 *	// some optional string<br>
	 *	String optString;<br>
	 *	...<br>
	 * <br>
	 * 	public void writeExternal(ObjectOutput out) throws IOException {<br>
	 * 		super.writeExternal(out);<br>
	 * <br>
	 * 		// write ALL custom message members, checking for NULL values where necessary, for example:<br>
	 * <br>
	 * 		// write the integer<br>
	 * 		out.writeInt(this.getSomeInt());<br>
	 * <br>
	 * 		// write the required string<br>
	 * 		out.writeUTF(this.getReqdString());<br>
	 *<br> 
	 * 		// and write the optional string, if it is defined<br>
	 * 		final boolean hasOptString = this.getOptString() == null;<br>
	 * 		out.writeBoolean(hasOptString);<br>
	 * 		if (hasOptString) {<br>
	 * 			out.writeUTF(this.getOptString());<br>
	 * 		}<br>
	 *  }<br>
	 *  }<br> 
	 */
	public void writeExternal(ObjectOutput out) throws IOException {
		// write the service name
		out.writeUTF(this.getServiceName());
		
		// write the connector ref
		boolean hasAttr = this.getConnectorRef() != null;
		out.writeBoolean(hasAttr);
		if (hasAttr) {
			out.writeUTF(this.getConnectorRef());
		}
		
		// finally, if we have a sender host defined, write the host and port
		hasAttr = this.getSenderHost() != null;
		out.writeBoolean(hasAttr);
		if (hasAttr) {
			// and write the sender host
			out.writeUTF(this.getSenderHost());
			
			// and port
			out.writeInt(this.getSenderPort());
		}
		
		// if we have an operation name, write it
		hasAttr = this.getOpName() != null;
		out.writeBoolean(hasAttr);
		if (hasAttr) {
			out.writeUTF(this.getOpName());
		}
	}
	
	/**
	 * See the java.io.Externalizable interface.
	 * 
	 * An example of how to implement this methods is given below, for a sample custom message class, MyMessage.<br>  
	 * The example below assumes that the writeExternal method is implemented as described in the documentation for that method in this class.
	 * <br>
	 * <br>
	 * 
	 * <code>
	 * public class MyMessage extends AbstractMessage {<br>
	 * <br>
	 * 	...<br>
	 *	// custom message members and methods, for example:<br>
	 *<br>
	 *	// some integer<br>
	 *	int someInt;<br>
	 *<br>
	 *	// some required string<br>
	 *	String reqdString;<br>
	 *<br>
	 *	// some optional string<br>
	 *	String optString;<br>
	 *	...<br>
	 *<br> 
	 *  public void readExternal(ObjectInput in) throws IOException {<br>
	 * 		super.readExternal(in);<br>
	 *<br> 
	 * 		// read ALL custom message members, checking for NULL values where necessary, for example:<br>
	 *<br> 
	 * 		// read the integer<br>
	 * 		this.setSomeInt(in.readInt());<br>
	 *<br> 
	 * 		// write the required string<br>
	 * 		this.setReqdString(in.readUTF());<br>
	 *<br> 
	 * 		// and write the optional string, if it is defined<br>
	 * 		final boolean hasOptString = in.readBoolean();<br>
	 * 		if (hasOptString) {<br>
	 * 			this.setOptString(in.readUTF());<br>
	 * 		}<br>
	 *  }<br>
	 * 
	 */
	public void readExternal(ObjectInput in) throws IOException {
		// read in the service name
		this.setServiceName(in.readUTF());		
		
		// read the connector ref
		boolean hasAttr = in.readBoolean();
		if (hasAttr) {
			this.setConnectorRef(in.readUTF());
		}
		
		// and then read in the sender host and port if we have one
		hasAttr = in.readBoolean();
		if (hasAttr) {
			
			// and then read the sender host
			this.setSenderHost(in.readUTF());
			
			// and sender port
			this.setSenderPort(in.readInt());
		}
		
		// see if we have an op name and if so read it
		hasAttr = in.readBoolean();
		if (hasAttr) {
			this.setOpName(in.readUTF());
		}
	}
	
	/**
     * Messages mostly know what types they are writing but we still need to encode user-supplied
     * objects e.g. parameters bound to prepared statements. Using this method of encoding objects
     * is much more efficient than calling writeObject(). However, this method will call writeObject
     * if it does not have an optimized way of writing a particular type.
     *
     * @param out output stream
     * @param obj object to write
     * @throws IOException if an i/o exception occurs
     */
    protected final void encodeObject(final ObjectOutput out, final Object obj) throws IOException {
        //TODO: add optimizations for all other primitive types
        if (obj instanceof Boolean) {
            out.writeByte( TYPE_BOOLEAN );
            out.writeBoolean((Boolean) obj);
        }
        else if (obj instanceof Character) {
            out.writeByte( TYPE_CHAR );
            out.writeChar((Character) obj);
        }
        else if (obj instanceof Short) {
            out.writeByte( TYPE_SHORT );
            out.writeShort((Short) obj);
        }
        else if (obj instanceof Integer) {
            out.writeByte( TYPE_INT );
            out.writeInt((Integer) obj);
        }
        else if (obj instanceof Long) {
            out.writeByte( TYPE_LONG );
            out.writeLong((Long) obj);
        }
        else if (obj instanceof Float) {
            out.writeByte( TYPE_FLOAT );
            out.writeFloat((Float) obj);
        }
        else if (obj instanceof Double) {
            out.writeByte( TYPE_DOUBLE );
            out.writeDouble((Double) obj);
        }
        else if (obj instanceof String) {
            out.writeByte( TYPE_STRING );
            out.writeUTF((String) obj);
        }
        else if (obj instanceof byte[]) {
            final byte[] data = (byte[]) obj;
            out.writeByte( TYPE_BYTEARRAY );
            out.writeInt( data.length );
            out.write(data, 0, data.length);
        }
        else {
            // fall back to serialization
            out.writeByte( TYPE_OBJECT );
            out.writeObject( obj );
        }
    }

    protected final Object decodeObject(final ObjectInput in) throws IOException, ClassNotFoundException {
        final byte type = in.readByte();
        switch (type) {
            case TYPE_BOOLEAN:
                return in.readBoolean();
            case TYPE_CHAR:
                return in.readChar();
            case TYPE_SHORT:
                return in.readShort();
            case TYPE_INT:
                return in.readInt();
            case TYPE_LONG:
                return in.readLong();
            case TYPE_FLOAT:
                return in.readFloat();
            case TYPE_DOUBLE:
                return in.readDouble();
            case TYPE_STRING:
                return in.readUTF();
            case TYPE_BYTEARRAY:
                int length = in.readInt();
                byte data[] = new byte[length];
                int bytesRead = 0;
                while (bytesRead < length) {
                    int x = in.read( data, bytesRead, length-bytesRead );
                    if (x == -1) {
                        throw new IOException( "Hit EOF unexpectedly" );
                    }
                    bytesRead += x;
                }
                if (bytesRead != data.length) {
                    throw new IOException( "Error reading byte array. Expected " + length + " bytes but only found " + bytesRead + " bytes." );
                }
                return data;
            case TYPE_OBJECT:
            default:
                return in.readObject();
        }
    }
    
    /**
     * Checks for message content equality.
     */
    public boolean equals(Object obj) {
    	// check the type
    	if (!(obj instanceof AbstractMessage)) {
    		return false;
    	}
    	AbstractMessage msg = (AbstractMessage) obj;
    	
    	// check the connector ref
    	// TODO: check the host and port?  it shouldn't matter, as the connector that provided the connector ref has that info
    	if (!ConfigHelper.checkStringMemsEqual(this.getConnectorRef(), msg.getConnectorRef())) {
    		return false;
    	}
    	
    	// and then the destination service name
    	if (!ConfigHelper.checkStringMemsEqual(this.getServiceName(), msg.getServiceName())) {
    		return false;
    	}
    	
    	return true;
    }

    /**
     * Outputs a somewhat readable string representation of the message.
     */
	public String toString() {
        return "\n[Message]\n\tDestination Service Name:\t" + this.getServiceName();
    }
}
