package org.softwarepipelines.framework.nio;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.log4j.Logger;
import org.softwarepipelines.framework.connector.NIOPipelinesConnector;
import org.softwarepipelines.framework.connector.PipelinesConnectorFactory;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.util.MessageHelper;

public class ClientMessageBroker extends MessageBrokerBase {
	private final static Logger LOG = Logger.getLogger(ClientMessageBroker.class);
	
    /* Channel To Connector Map */
    protected HashMap<SocketChannel, NIOPipelinesConnector> scToConnectorMap;

    /* Connector UUID Ref to instance map */
    protected HashMap<String, NIOPipelinesConnector> connectorRefMap;
    
    /**
     * Creates a new ClientMessageBroker, listening for responses on a ServerSocketChannel configured according
     * to the provided socket config.
     */
    public ClientMessageBroker() throws PipelineException {
    	super();
    	
        // instantiate the callback map
        connectorRefMap = new HashMap<String, NIOPipelinesConnector>();
    }
    
    /**
     * Bind the listening server socket channel to the responseHost and responsePort
     * in the config, as client message brokers serve as service response destinations
     */
    public void bindListener() throws IOException {
    	ssc.socket().bind(new InetSocketAddress(socketConnConfig.getResponseHost(),
                socketConnConfig.getResponsePort()));
    }
    
    /**
     * Iterate over the valid keys provided by the selector.
     */
    public void run() {
        SelectionKey key = null;
        Iterator<SelectionKey> selKeys;

        // selection loop
        alive = true;

        while (alive) {
            try {
            	
            	// process pending changes
            	processPendingChanges();
            	
                // select keys
                this.getSelector().select();

                // iter over the selected set
                selKeys = this.getSelector().selectedKeys().iterator();

                while (selKeys.hasNext()) {
                    key = (SelectionKey) selKeys.next();
                    selKeys.remove();

                    // check validity
                    if (!key.isValid()) {
                        continue;
                    }

                    // if the key is acceptable, accept a connection in this thread
                    if (key.isAcceptable()) {
                        this.accept(key);
                    }
                    // readable
                    else if (key.isReadable()) {
                        this.read(key);
                    }
                    // writeable
                    else if (key.isWritable()) {
                        this.write(key);
                    }
                }
            } catch (Exception e) {
                LOG.warn("Error in selection process", e);
                key.cancel();
                return;
            }
        }
    }

    /**
     * Accept an incoming connection from a MessageSender; register the newly-connected SocketChannel with the selector.
     *
     * @param key the key for which ACCEPT ops are valid
     */
    private void accept(SelectionKey key) throws IOException, PipelineException {
        // Get the server socket channel
        ServerSocketChannel ssc = (ServerSocketChannel) key.channel();

        // and try to accept a connection
        if (ssc.isOpen()) {
     
        	// accept the connection
        	SocketChannel sc = ssc.accept();
            sc.configureBlocking(false);
            
            // and enqueue a change request
            OpsChangeRequest change = new OpsChangeRequest();
            change.setAttachment(ByteBuffer.allocate(2048));
            change.setChannel(sc);
            change.setOps(SelectionKey.OP_READ);
            change.setType(OpsChangeRequest.REGISTER_CHANGE_TYPE);
            while (!pendingChanges.contains(change)) {
            	try {
            		pendingChanges.put(change);
            		break;
            	} catch (InterruptedException ie) {
            		LOG.warn("Interruped putting change request on queue");
            	}
            }

            // log the new connection
            activeConns++;
            LOG.info("ClientMessageBroker accepted new connection, now has " + activeConns);
            
        } else {
            LOG.warn(
                "Trying to accept a connection over a closed SSC, no connection established!");
        }
    }

    /**
     * Read an incoming message from the SocketChannel and use the appropriate router to route the message.
     *
     * @param key the key for which READ ops are valid
     */
    private void read(SelectionKey key) throws IOException, PipelineException {
        // Get the SocketChannel from the key
        SocketChannel sc = (SocketChannel) key.channel();

        // and then get the read buffer attachment
        ByteBuffer readBuf = (ByteBuffer) key.attachment();

        // Read the message through the channel into the read buffer
        int read = 0;
        try {
        	read = sc.read(readBuf);
        } catch (IOException ioe) {
        	if (!sc.isConnected()) {
        		activeConns--;
        		this.getScToConnectorMap().remove(sc);
        		LOG.warn("Error reading dis-connected socket channel, removing callback map entry, active conns: " + activeConns);
        		return;
        	}
        }
        
        //
        if (read == 0 || read == readBuf.capacity()) {
        	readBuf.clear();
        	return;
        }
        
        readBuf.flip();

        // Create the AbstractMessage(s) from the read buffer
        AbstractMessage incomingMsg;
        NIOPipelinesConnector conn;
        while (readBuf.hasRemaining()) {
        	incomingMsg = MessageHelper.readMessageFromBuffer(readBuf, this.getDistConfig().getServiceToMsgClassMap());
        	LOG.debug("Read message from buffer: " + incomingMsg);
        	
        	// if the incoming message is null, we had an incomplete read; return as the buffer is ready to complete the read
        	if (incomingMsg == null) {
        		return;
        	}
        	
        	// read the message's connector reference, and look it up in the connector reference map
        	if (!getConnectorRefMap().containsKey(incomingMsg.getConnectorRef())) {
        		throw new PipelineException("No connector with UUID ref (" + incomingMsg.getConnectorRef() + 
        				") registered with ClientMessageBroker for distributor " + this.getDistConfig().getName());
        	}
        	conn = getConnectorRefMap().get(incomingMsg.getConnectorRef());
        	
        	// try to put the incoming message on the connector's incoming queue
        	do {
        		try {
        			conn.getIncomingMsgs().put(incomingMsg);
        			break;
        		} catch (InterruptedException ie) {
        			LOG.warn("Interrupted enqueueing incoming message, trying again: " + incomingMsg);
        		}
        	} while (!conn.getIncomingMsgs().contains(incomingMsg));
        }
        
        // compact the read buffer after a clean message read
        readBuf.compact();
    }

    /**
     * Write
     *
     * @param key the key for which WRITE ops are valid
     */
    private void write(SelectionKey key) throws IOException {
        // Get the SocketChannel from the key
        SocketChannel sc = (SocketChannel) key.channel();

        // and then get the read buffer attachment
        ByteBuffer writeBuf = (ByteBuffer) key.attachment();
        
        // if no write buffer attached, log the error and cancel the key 
        if (writeBuf == null ) {
        	LOG.error("Null write buffer attached to key, cancelling key!");
        	key.cancel();
        	return;
        }
        
        // synchronize on the write buffer in case we are trying to write the next message into the buffer
        synchronized (writeBuf) {
        	
        	// flip the write buffer to prepare for gets
        	writeBuf.flip();
        	
            // Write the buffer's contents through the channel
            while (writeBuf.hasRemaining()) {
                try {
                    sc.write(writeBuf);
                } catch (IOException ioe) {
                	LOG.warn("Error writing to channel, closing...");
                	sc.close();
            		key.cancel();
            		return;
                }
            }
            
            // clear the write buffer after a successful write
            writeBuf.clear();

		}

        // set the empty interest ops on the key, as the next call through the connector will have to tell us it is 
        //ready (otherwise we just return to write again and again, with nothing in the buffer)
        OpsChangeRequest change = new OpsChangeRequest();
    	change.setAttachment(writeBuf);
    	change.setChannel(sc);
    	change.setType(OpsChangeRequest.OP_CHANGE_TYPE);
    	change.setOps(0);
    	enqueueChangeRequest(change);
    }
    
    /**
     * Checks out an NIOPipelinesConnector.
     * 
     * The connector is not yet connected upon checkout, but will connect when a message is sent (if not connected before).
     * 
     * @return the socket connector
     * @throws PipelineException
     */
    public synchronized NIOPipelinesConnector checkoutSocketConnector() throws PipelineException {
    	// build the connector in the factory
    	NIOPipelinesConnector conn = (NIOPipelinesConnector) PipelinesConnectorFactory.getPipelinesConnector(socketConnConfig);
    	conn.setBroker(this);
    	
    	// connect the connector to the other broker
    	//conn.connect();
    	
    	// register the connector in both connector maps
    	connectorRefMap.put(conn.getConnectorRef().toString(), conn);
    	
    	// finally, return the connector
    	return conn;
    }
    
    /**
     * Checks out an NIOPipelinesConnector.
     * 
     * The connector is not yet connected upon checkout, but will connect when a message is sent (if not connected before).
     * 
     * @return the socket connector
     * @throws PipelineException
     */
    public synchronized void checkinSocketConnector(final NIOPipelinesConnector connector) {
    	// remove the connector ref in the map
    	connectorRefMap.remove(connector.getConnectorRef().toString());
    	
    	// if we have no connectors checked out, stop the task in the brokerExec
    	if (connectorRefMap.isEmpty()) {
    		// remove the message broker from the factory
    		MessageBrokerFactory.shutdownBroker(this);
    		
    		// and shut it down
    		this.shutdown();
    	}
    }

	public HashMap<SocketChannel, NIOPipelinesConnector> getScToConnectorMap() {
		return scToConnectorMap;
	}

	public void setScToConnectorMap(
			HashMap<SocketChannel, NIOPipelinesConnector> scToConnectorMap) {
		this.scToConnectorMap = scToConnectorMap;
	}

	public HashMap<String, NIOPipelinesConnector> getConnectorRefMap() {
		return connectorRefMap;
	}

	public void setConnectorRefMap(
			HashMap<String, NIOPipelinesConnector> connectorRefMap) {
		this.connectorRefMap = connectorRefMap;
	}
}
