package org.softwarepipelines.framework.nio;

import java.io.IOException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.spi.SelectorProvider;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.util.ConfigHelper;
import org.softwarepipelines.framework.exception.PipelineException;

/**
 * Base class for socket based communication message brokers.
 * 
 * @author cholm
 *
 */
public abstract class MessageBrokerBase implements Runnable {
	final private static Logger LOG = Logger.getLogger(MessageBrokerBase.class);
	
	/* Executor; thread executor in which this broker will run */
	protected ExecutorService brokerExec;
	
	/* Selector */
    protected Selector selector;

    /* Server Socket Channel for accepting connections */
    protected ServerSocketChannel ssc;

    /* Distributor Config */
    protected PipelinesDistributorConfig distConfig;
    
    /* Socket Connector Config */
    protected PipelinesConnectorConfig socketConnConfig;
    
    /* Pending Changes Queue */
    protected ArrayBlockingQueue<OpsChangeRequest> pendingChanges;
    
    /* Alive Flag */
    protected boolean alive = false;

    /* Active Connections */
    protected int activeConns;
    
    /**
     * 
     * @throws PipelineException
     */
    public MessageBrokerBase() throws PipelineException {
		// set active connections to 0
	    activeConns = 0;
	
	    // instantiate the pending changes queue
	    pendingChanges = new ArrayBlockingQueue<OpsChangeRequest>(10);
	    
	    // and finally setup the broker thread executor
	    brokerExec = Executors.newSingleThreadExecutor();
    }
    
    /**
     * 
     * @param distConfig
     */
    public void configure(final PipelinesDistributorConfig aConfig) throws PipelineException {
    	// set the distributor config
	    distConfig = aConfig;
	    
	    // check that the distributor has a socket connector configured
	    // TODO: switch to ConfigHelper util class
	    socketConnConfig = ConfigHelper.getConnectorConfigByType(getDistConfig(), PipelinesConnectorConfig.SOCKET_CONNECTOR_TYPE);
	    if (socketConnConfig == null) {
	    	throw new PipelineException("No socket connector configured on PipelinesDistributor in MessageBroker!");
	    }
	
	     // open the selector
	    try {
	        selector = SelectorProvider.provider().openSelector();
	
	        // and open and bind the server socket channel to the send to host and port, as the MessageBroker receives request messages
	        // from clients or PipelineServices
	        ssc = selector.provider().openServerSocketChannel();
	        ssc.configureBlocking(false);
	        
	        // bind the server socket channel to the appropriate socket addr
		    bindListener();
	    } catch (IOException ioe) {
	        LOG.error("Exception setting up Selector and/or ServerSocketChannel!",
	            ioe);
	        throw new PipelineException("Exception setting up Selector and/or ServerSocketChannel!",
	            ioe);
	    }
	
	    // then register the channel for ACCEPT ops
	    try {
	        ssc.register(selector, SelectionKey.OP_ACCEPT);
	    } catch (ClosedChannelException cce) {
	        LOG.error("Exception registering ServerSocketChannel for ACCEPT ops!",
	            cce);
	        throw new PipelineException("Exception registering ServerSocketChannel for ACCEPT ops!",
	            cce);
	    }
    }

    /**
     * Sub classes must implement the selection and I/O operation logic in the run() method. 
     */
    public abstract void run();
    
    /**
     * 
     */
    public abstract void bindListener() throws IOException;
    
    /**
     * Enqueue an ops change request.
     * @param change
     */
    public synchronized void enqueueChangeRequest(final OpsChangeRequest change) {
    	while (!pendingChanges.contains(change)) {
    		try {
    			pendingChanges.put(change);
    			break;
    		} catch (InterruptedException ie) {
    			LOG.warn("Interrupted enqueuing change: " + change, ie);
    		}
    	}
    }
    
    /**
     * Processes the ops change requests in the pending changes queue.
     * @throws PipelineException
     */
    protected void processPendingChanges() throws PipelineException {
    	OpsChangeRequest change;
    	SelectionKey key;
    	
    	// process any pending changes
    	while (!pendingChanges.isEmpty()) {
    		try {
	    		change = pendingChanges.take();
	    		
	    		// TODO: check for invalid key
	    		
	    		// register
	    		if (change.getType() == OpsChangeRequest.REGISTER_CHANGE_TYPE) {
	    			key = change.getChannel().register(selector, change.getOps());
	    			key.attach(change.getAttachment());
	    		}
	    		// change
	    		else if (change.getType() == OpsChangeRequest.OP_CHANGE_TYPE) {
	    			key = change.getChannel().keyFor(selector);
	    			key.interestOps(change.getOps());
	    			key.attach(change.getAttachment());
	    		} else {
	    			throw new PipelineException("Invalid change request type: " + change.getType());
	    		}
    		} catch (InterruptedException ie) {
    			LOG.warn("Interrupted while trying to take the next pending change!");
    		} catch (ClosedChannelException cce) {
    			throw new PipelineException("Attempted to register closed channel!");
    		}
    	}
    }
    
    /**
     * Shut down this distributor.
     */
    public synchronized void shutdown() {
    	if (!alive) {
    		return;
    	}
    	// set alive to false
    	alive = false;
    	
    	// wake up the selector
    	this.getSelector().wakeup();
    	
    	// and wait for any valid keys to process; we will hop out afterwards
    	try {Thread.sleep(1000);} catch (Exception e){}
    	
    	// and try to close the selector and server socket channel
        try {
        	getSelector().close();
            ssc.close();
            brokerExec.shutdownNow();
        } catch (IOException ioe) {
            LOG.warn("Exception closing ServerSocketChannel!", ioe);
        }
        
        // and finally, remove ourselves from the factory
        MessageBrokerFactory.shutdownBroker(this);
    }
    
    /**
     * If the message broker is not alive, kick it off in the thread executor.
     */
    public synchronized void start() {
    	//brokerExec.shutdownNow();
    	if (!this.isAlive()) {
    		brokerExec.execute(this);
    	}
    }

    /* GETTERS AND SETTERS */
	public ExecutorService getBrokerExec() {
		return brokerExec;
	}

	public void setBrokerExec(ExecutorService brokerExec) {
		this.brokerExec = brokerExec;
	}

	public Selector getSelector() {
		return selector;
	}

	public void setSelector(Selector selector) {
		this.selector = selector;
	}

	public ServerSocketChannel getSsc() {
		return ssc;
	}

	public void setSsc(ServerSocketChannel ssc) {
		this.ssc = ssc;
	}

	public PipelinesDistributorConfig getDistConfig() {
		return distConfig;
	}

	public void setDistConfig(PipelinesDistributorConfig distConfig) {
		this.distConfig = distConfig;
	}

	public PipelinesConnectorConfig getSocketConnConfig() {
		return socketConnConfig;
	}

	public void setSocketConnConfig(PipelinesConnectorConfig socketConnConfig) {
		this.socketConnConfig = socketConnConfig;
	}

	public ArrayBlockingQueue<OpsChangeRequest> getPendingChanges() {
		return pendingChanges;
	}

	public void setPendingChanges(
			ArrayBlockingQueue<OpsChangeRequest> pendingChanges) {
		this.pendingChanges = pendingChanges;
	}

	public boolean isAlive() {
		return alive;
	}

	public void setAlive(boolean alive) {
		this.alive = alive;
	}

	public int getActiveConns() {
		return activeConns;
	}

	public void setActiveConns(int activeConns) {
		this.activeConns = activeConns;
	}
}
