package org.softwarepipelines.framework.nio;

import org.apache.log4j.Logger;

import org.softwarepipelines.framework.callback.PipelinesCallbackHandler;
import org.softwarepipelines.framework.distributor.PipelinesDistributorBase;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.util.MessageHelper;

import java.io.IOException;

import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;

import java.util.HashMap;
import java.util.Iterator;

/**
 * Simple Push Pattern Distributor;
 *
 * @author cholm
 */
public class ServerMessageBroker extends MessageBrokerBase {
    private final static Logger LOG = Logger.getLogger(ServerMessageBroker.class);

    /* Message Distributor */
    protected PipelinesDistributorBase distributor;

    /* Message Broker Callback Handler Map */
    protected HashMap<SocketChannel, PipelinesCallbackHandler> connToCallbackMap;    
    
    /**
     *
     */
    public ServerMessageBroker() throws PipelineException {
    	super();

        // and instantiate the callback map
        connToCallbackMap = new HashMap<SocketChannel, PipelinesCallbackHandler>();
    }
    
    /**
     * Bind the listening server socket channel to the sendToHost and sendToPort
     *  in the config, as server message brokers serve as service request destinations
     */
    public void bindListener() throws IOException {
    	ssc.socket().bind(new InetSocketAddress(socketConnConfig.getSendToHost(),
                socketConnConfig.getSendToPort()));
    }

    /**
     * Iterate over the valid keys provided by the selector.
     */
    public void run() {
        SelectionKey key = null;
        Iterator<SelectionKey> selKeys;

        // selection loop
        alive = true;

        while (alive) {
            try {
            	
            	// process pending changes
            	processPendingChanges();
            	
                // select keys
                this.getSelector().select();

                // iter over the selected set
                selKeys = this.getSelector().selectedKeys().iterator();

                while (selKeys.hasNext()) {
                    key = (SelectionKey) selKeys.next();
                    selKeys.remove();

                    // check validity
                    if (!key.isValid()) {
                        continue;
                    }

                    // if the key is acceptable, accept a connection in this thread
                    if (key.isAcceptable()) {
                        this.accept(key);
                    }
                    // readable
                    else if (key.isReadable()) {
                        this.read(key);
                    }
                    // writeable
                    else if (key.isWritable()) {
                        this.write(key);
                    }
                }
            } catch (Exception e) {
                LOG.warn("Error in selection process", e);
                key.cancel();
            }
        }
    }

    /**
     * Accept an incoming connection from a MessageSender; register the newly-connected SocketChannel with the selector.
     *
     * @param key the key for which ACCEPT ops are valid
     */
    private void accept(SelectionKey key) throws IOException, PipelineException {
        // Get the server socket channel
        ServerSocketChannel ssc = (ServerSocketChannel) key.channel();

        // and try to accept a connection
        if (ssc.isOpen()) {
            
        	SocketChannel sc = ssc.accept();
            sc.configureBlocking(false);
            
            // and enqueue a change request
            OpsChangeRequest change = new OpsChangeRequest();
            change.setAttachment(ByteBuffer.allocate(2048));
            change.setChannel(sc);
            change.setOps(SelectionKey.OP_READ);
            change.setType(OpsChangeRequest.REGISTER_CHANGE_TYPE);
            enqueueChangeRequest(change);

            // log the new connection
            activeConns++;
            LOG.info("MessageBroker accepted new connection, now has " + activeConns);
            
        } else {
            LOG.warn(
                "Trying to accept a connection over a closed SSC, no connection established!");
        }
    }

    /**
     * Read an incoming message from the SocketChannel and use the appropriate router to route the message.
     *
     * @param key the key for which READ ops are valid
     */
    private void read(SelectionKey key) throws IOException, PipelineException {
        // Get the SocketChannel from the key
        SocketChannel sc = (SocketChannel) key.channel();

        // and then get the read buffer attachment
        ByteBuffer readBuf = (ByteBuffer) key.attachment();

        // Read the message through the channel into the read buffer
        try {
        	sc.read(readBuf);
        } catch (IOException ioe) {
        	if (!sc.isConnected()) {
        		activeConns--;
        		this.getConnToCallbackMap().remove(sc);
        		LOG.warn("Error reading dis-connected socket channel, removing callback map entry, active conns: " + activeConns);
        		return;
        	}
        }
        if (readBuf.position() == 0 && readBuf.limit() == readBuf.capacity()) {
        	//LOG.warn("Nothing read!");
        	//key.cancel();
        	return;
        }
        
        // flip the read buffer to prepare for reading the message
        readBuf.flip();

        // Create the AbstractMessage(s) from the read buffer
        AbstractMessage incomingMsg;
        while (readBuf.hasRemaining()) {
        	incomingMsg = MessageHelper.readMessageFromBuffer(readBuf, this.getDistConfig().getServiceToMsgClassMap());
        	LOG.debug("Read message from buffer: " + incomingMsg);
        	
        	// if the incoming message is null, we had an incomplete read; return as the buffer is ready to complete the read
        	if (incomingMsg == null) {
        		LOG.debug("returning after reading null msg");
        		return;
        	}
        
	        // and pass it to the PipelinesDistributor
	        distributor.distribute(incomingMsg);
        }
        
        // compact the read buffer after a clean message read
        readBuf.clear();
    }

    /**
     * Write
     *
     * @param key the key for which WRITE ops are valid
     */
    private void write(SelectionKey key) throws IOException {
        // Get the SocketChannel from the key
        SocketChannel sc = (SocketChannel) key.channel();

        // and then get the read buffer attachment
        ByteBuffer writeBuf = (ByteBuffer) key.attachment();

        // Write the buffer's contents through the channel
        int tries = 0;
        writeBuf.flip();
        while (writeBuf.hasRemaining()) {
            try {
                sc.write(writeBuf);
            } catch (IOException ioe) {
                LOG.warn("Error writing to channel, retrying...");
                tries++;

                try {
                    Thread.sleep(100);
                } catch (Exception e) {
                }
            }
        }

        writeBuf.clear();

        // finally, register the channel for READ ops to receive the next message(s)
        OpsChangeRequest change = new OpsChangeRequest();
        change.setType(OpsChangeRequest.OP_CHANGE_TYPE);
        change.setOps(SelectionKey.OP_READ);
        change.setChannel(sc);
        change.setAttachment(writeBuf);
        enqueueChangeRequest(change);
    }

	public PipelinesDistributorBase getDistributor() {
		return distributor;
	}

	public void setDistributor(PipelinesDistributorBase distributor) {
		this.distributor = distributor;
	}

	public HashMap<SocketChannel, PipelinesCallbackHandler> getConnToCallbackMap() {
		return connToCallbackMap;
	}

	public void setConnToCallbackMap(
			HashMap<SocketChannel, PipelinesCallbackHandler> connToCallbackMap) {
		this.connToCallbackMap = connToCallbackMap;
	}
}
