package org.softwarepipelines.framework.pipeline;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.pipeline.PipelineConfig;
import org.softwarepipelines.config.util.ConfigHelper;
import org.softwarepipelines.framework.distributor.PipelinesDistributorBase;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.service.PipelineServiceBase;
import org.softwarepipelines.framework.service.PipelineServiceFactory;
import org.softwarepipelines.framework.service.RelayService;
import org.softwarepipelines.framework.util.Counter;

public class PipelineImpl implements Pipeline {
	final private static Logger LOG = Logger.getLogger( PipelineImpl.class );
	
	/* Config */
	protected PipelineConfig config;
	
	/* Pipeline Service to invoke */
	protected PipelineServiceBase service;
	
	/* Distributor Reference */
	protected PipelinesDistributorBase distributor;
	
	/* Executor */
	protected ExecutorService executor;
	
	/* AbstractMessage Queues */
	protected ArrayBlockingQueue<AbstractMessage> inQ;
	
	/* Counter */
	protected Counter counter;

	public PipelineImpl() {
	}
	
	public PipelineImpl(final PipelineConfig aConfig, final PipelinesDistributorBase distributor, final Counter counter) throws PipelineException {
		this.setDistributor(distributor);
		this.setCounter(counter);
		configure(aConfig);
	}

	public void configure(final PipelineConfig aConfig) throws PipelineException {
		// set the config
		config = aConfig;
		
		// instantiate the incoming message queue
		inQ = new ArrayBlockingQueue<AbstractMessage>(20);//PipelineConfig.DEF_INQ_CAP);
		
		// see if we are relaying, if so check out the connector
		if (config.getType().equals(PipelineConfig.RELAY_PIPELINE_TYPE)) {
			LOG.debug("Configuring relay pipeline...");
			if (config.getConnConfig().getType().equals(PipelinesConnectorConfig.LOCAL_CONNECTOR_TYPE)) {
				
				throw new PipelineException("Relay pipelines must reference a non-local connector in the send-to distributor");
				
			} else if (config.getConnConfig().getType().equals(PipelinesConnectorConfig.SOCKET_CONNECTOR_TYPE)){
				
				service = PipelineServiceFactory.getSocketRelayService(config);
				
			// TODO: add ws connectors
			} else {
				throw new PipelineException("Unsupported relay connector type: " + config.getConnConfig().getType());
			}
		}
		// otherwise we are a local pipeline, so load an instance of the java class
		else {
			service = PipelineServiceFactory.getPipelineService(config.getServiceConfig());
		}
		
		// set the incoming message queue, callback reference map, instrumentor counter and pipeline name
		service.setInQ(inQ);
		service.setDistributor(this.getDistributor());
		service.setCounter(this.getCounter());
		service.setPipelineName(this.getConfig().getName());
		
		// if we have a relay service, init the connector
		try {
			if (ConfigHelper.isDistributorServiceRelay(service.getConfig())) {
				if (service instanceof RelayService) {
					((RelayService)service).initRelay();
				}
			}
		} catch (ConfigException ce) {
			throw new PipelineException("Error checking if service " + service.getName() + " is relay type", ce);
		}
		
		// and finally, build the executor
		buildExecutor();

	}
	
	/**
	 * Build the Executor in which the pipeline will execute tasks.
	 * 
	 * Pipelines associated with content based routers should be configured to have a single thread executor; those
	 * associated with round robin routers may define a min and max threads value and we will build a thread pool executor
	 * with a discard policy.  The discard policy will not result in dropped messages, as the messages will already have been
	 * enqueued, and all PipelineServices clear the entire incoming message queue before they are finished executing as runnable
	 * tasks.
	 * 
	 * This means that we can get into situations rather easily where each message coming in has an appropriate task built and automatically
	 * discarded because a previous task has yet to handle all messages in the queue.  Perhaps we should look into finding a way to determine
	 * when we should and should not kick a new task (that is, check if a task is currently running or not, I think).
	 */
	private void buildExecutor() {
	
		// XXX: for now, we only support single threaded, versus thread pool based, pipelines
		executor = (ExecutorService) Executors.newSingleThreadExecutor();
	}

	public void invokeService(AbstractMessage msg) throws PipelineException {
		// enqueue the message for processing
		enqueueMessage(msg);
		
		// and then kick off a service task (if we need to)
		executor.execute(service);
	}
	
	private void enqueueMessage(final AbstractMessage msg) throws PipelineException {
		 do {
	            try {
	                inQ.put(msg);
	                LOG.debug("Pipeline " + this.getConfig().getName() + " now has " + inQ.size() + " messages in queue");
	                break;
	            } catch (InterruptedException ie) {
	                LOG.warn("Interrupted while enqueueing message: " + msg +
	                    "\nRetrying enqueue...");
	            }
	        } while (!inQ.contains(msg));
	}
	
	public PipelineConfig getConfig() {
		return config;
	}

	public void setConfig(PipelineConfig config) {
		this.config = config;
	}

	public ExecutorService getExecutor() {
		return executor;
	}

	public void setExecutor(ExecutorService executor) {
		this.executor = executor;
	}

	public ArrayBlockingQueue<AbstractMessage> getInQ() {
		return inQ;
	}

	public void setInQ(ArrayBlockingQueue<AbstractMessage> inQ) {
		this.inQ = inQ;
	}

	public Counter getCounter() {
		return counter;
	}

	public void setCounter(Counter counter) {
		this.counter = counter;
	}

	public PipelineServiceBase getService() {
		return service;
	}

	public void setService(PipelineServiceBase service) {
		this.service = service;
	}

	public PipelinesDistributorBase getDistributor() {
		return distributor;
	}

	public void setDistributor(PipelinesDistributorBase distributor) {
		this.distributor = distributor;
	}
}
