package org.softwarepipelines.framework.router;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;

import org.softwarepipelines.config.router.PipelinesRouterConfig;
import org.softwarepipelines.config.pipeline.PipelineConfig;
import org.softwarepipelines.framework.distributor.PipelinesDistributorBase;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.pipeline.Pipeline;
import org.softwarepipelines.framework.pipeline.PipelineFactory;
import org.softwarepipelines.framework.pipeline.PipelineImpl;
import org.softwarepipelines.framework.util.Counter;
import org.softwarepipelines.framework.util.Instrumentor;

public abstract class PipelinesRouterBase implements PipelinesRouter {
	
	/* Router Config */
	protected PipelinesRouterConfig config;
	
	/* Distributor Reference */
	protected PipelinesDistributorBase distributor;
	
	/* Pipelines List */
	protected List<Pipeline> pipelines;
	
	/* Instrumentor */
	private Instrumentor instrumentor;
	
	public PipelinesRouterBase() {
	}
	
	public PipelinesRouterBase(final PipelinesRouterConfig aConfig, final PipelinesDistributorBase dist) throws PipelineException {
		this.setDistributor(dist);
		configure(aConfig);
	}

	/**
	 * Configure the PipelinesRouter.
	 * 
	 * Sets the config and uses the PipelinesRouterFactory to build the Pipelines list.
	 */
	public void configure(PipelinesRouterConfig aConfig) throws PipelineException {
		List<PipelineConfig> plConfigs;
		PipelineImpl pipeline;
		Counter counter;
		
		// set the config
		config = aConfig;
		
		// get the instrumentor
		instrumentor = Instrumentor.getInstrumentor(PipelinesRouterBase.class);
		
		// and then build the pipelines list
		this.pipelines = new ArrayList<Pipeline>();
		plConfigs = config.getServiceConfig().getPipelines();
		for (PipelineConfig plConfig : plConfigs) {
			counter = instrumentor.getCounter(plConfig.getName());
			pipeline = PipelineFactory.getPipeline(plConfig, this.getDistributor(), counter);
			pipelines.add(pipeline);
		}
	}
	
	protected Object getPipelineKey(final AbstractMessage msg) throws PipelineException {
		Class<?> msgClass;
		Field keyField;
		
		// get the class 
		if (!getDistributor().getConfig().getServiceToMsgClassMap().containsKey(msg.getServiceName())) {
			throw new PipelineException("Couldn't find message class for service : " + msg.getServiceName() + 
					" in map: " + getDistributor().getConfig().getServiceToMsgClassMap());
		}
		msgClass = getDistributor().getConfig().getServiceToMsgClassMap().get(msg.getServiceName());
		
		// and then look for the pipeline key field name
		try {
			keyField = msgClass.getField(config.getMessageConfig().getPipelineKey());
		} catch (NoSuchFieldException nsfe) {
			throw new PipelineException("No public field " + config.getMessageConfig().getPipelineKey() + 
					" on Message class " + msgClass.getName(), nsfe);
		} catch (NullPointerException npe) {
			throw new PipelineException("No pipeline key specified in message config for message type: " + config.getMessageConfig().getType());
		}
		
		// return the value of the field on the given message
		try {
			return keyField.get(msg);
		} catch (IllegalAccessException iae) {
			throw new PipelineException("Error getting field " + keyField.getName() + " on Message casted to " + msgClass.getName(), iae);
		}  
	}
	
	/**
	 * Iterate through the pipelines and shut down each executor.
	 */
	public void shutdown() throws PipelineException {
		for (Pipeline pl : this.getPipelines()) {
			((PipelineImpl)pl).getExecutor().shutdownNow();
			
			// shut down the pipeline's service
			((PipelineImpl)pl).getService().shutdown();
		}
	}

	/**
	 * Sub-classes of the base router class implement the routing logic in route(AbstractMessage). 
	 */
	public abstract void route(AbstractMessage msg) throws PipelineException;

	public PipelinesRouterConfig getConfig() {
		return config;
	}

	public void setConfig(PipelinesRouterConfig config) {
		this.config = config;
	}

	public List<Pipeline> getPipelines() {
		return pipelines;
	}

	public void setPipelines(List<Pipeline> pipelines) {
		this.pipelines = pipelines;
	}

	public PipelinesDistributorBase getDistributor() {
		return distributor;
	}

	public void setDistributor(PipelinesDistributorBase distributor) {
		this.distributor = distributor;
	}
}
