package org.softwarepipelines.framework.service;

import org.apache.log4j.Logger;

import org.softwarepipelines.config.PipelinesMasterConfig;
import org.softwarepipelines.config.service.PipelineServiceConfig;
import org.softwarepipelines.framework.callback.PipelinesCallbackHandler;
import org.softwarepipelines.framework.callback.PipelinesSocketCallbackHandlerImpl;
import org.softwarepipelines.framework.distributor.PipelinesDistributorBase;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.util.Counter;

import java.util.concurrent.ArrayBlockingQueue;

public abstract class PipelineServiceBase implements PipelineService, Runnable {
    private static final Logger LOG = Logger.getLogger(PipelineServiceBase.class);
    
    /* Current Message Reference */
    protected AbstractMessage currMsg;
    
    /* Parent Distributor Reference */
    protected PipelinesDistributorBase distributor;
    
    /* Service Config */
    protected PipelineServiceConfig config;
    
    /* Service Name */
    protected String name;
    
    /* Pipeline Name */
    protected String pipelineName;
    
    /* Incoming Message Queues */
    protected ArrayBlockingQueue<AbstractMessage> inQ;
    
    /* Instrumentor Counter */
    protected Counter counter;

	public PipelineServiceBase() {
    }

    public PipelineServiceBase(final PipelineServiceConfig aConfig) {
        config = aConfig;
        name = aConfig.getName();
    }
    
    /**
     * All sub classes must override this method to process the incoming message (optionally calling sendResultMessage to send a response)
     */
    public abstract void processMessage(AbstractMessage msg) throws PipelineException;
    
    /**
     * Shutdown method; custom services override this method to implement clean up logic.
     * 
     * At the base class level, there is presently nothing to do.
     */
    public void shutdown() throws PipelineException {
    	LOG.debug("Shutting down service " + this.getName() + "...");
    }

    /**
     * Process all messages in the queue.
     */
    public void run() {
        
    	while (!inQ.isEmpty()) {
            try {
            	// take the message from the front of the queue
                currMsg = inQ.take();
                
                // and process the message
                this.processMessage(currMsg);
                
                // finally, invoke the counter
                counter.next();
                
            } catch (PipelineException pe) {
                LOG.error("Error processing message: " + currMsg, pe);
            } catch (InterruptedException ie) {
                LOG.warn("Interrupted taking message from queue, retrying...",
                    ie);
            }
        }
    }
    
    public void sendResultMessage(final AbstractMessage result) throws PipelineException {
    	PipelinesCallbackHandler callback;
    	// TODO: implement WS result messages

    	// set the destination information from the current message (the incoming message for which this is a response)
    	result.setConnectorRef(currMsg.getConnectorRef());
		result.setSenderHost(currMsg.getSenderHost());
		result.setSenderPort(currMsg.getSenderPort());
		result.setServiceName(currMsg.getServiceName());
    	
    	// check the connector reference to retrieve the callback handler
    	if (result.getConnectorRef() != null) {
    		
    		// if we have an entry in the callback map, call returnResult
    		if (getDistributor().getConnectorRefToCallbackMap().containsKey(result.getConnectorRef())) {
    			callback = getDistributor().getConnectorRefToCallbackMap().get(result.getConnectorRef());
    		}  
    		// otherwise, we should have a result message that we need to send to through
    		// a socket channel that is not currently open or connected
    		else {
    			
    			// validate that we have a host and port
    			if (result.getSenderHost() != null && result.getSenderPort() != -1) {
    				
    				callback = createSocketCallback(result.getSenderHost(), result.getSenderPort());
    				getDistributor().getConnectorRefToCallbackMap().put(result.getConnectorRef(), callback);
    			
    			} else {
    				throw new PipelineException("Error setting up socket connector, need to set a valid host and port!\thost: " + 
    						result.getSenderHost() + ", port: " + result.getSenderPort());
    			}
    		}
    		
    		// and finally, use the callback to return the result
    		callback.returnResult(result);
    		
    		
    	} else {
    		
    		// throw an exception, as we have no clue to which connector we should respond
    		throw new PipelineException("Can't send result message, no connector reference!\n" + result);
    	}
    }
    
    private PipelinesCallbackHandler createSocketCallback(final String host, final int port) throws PipelineException {
    	PipelinesSocketCallbackHandlerImpl callback;
    	
    	// validate that we have a broker on our distributor
    	if (getDistributor().getServerBroker() == null) {
    		throw new PipelineException("No broker reference on distributor " + getDistributor().getConfig().getName() + 
    				", can't create socket callback handler in service " + this.getName());
    	}
    	
    	// create a new callback handler
    	callback = new PipelinesSocketCallbackHandlerImpl();
    	callback.connectAndReg(host, port, getDistributor().getServerBroker().getSelector(), getDistributor().getServerBroker().getPendingChanges());
    	
    	// and return 
    	return callback;
    }
    
    protected String getDestMapKey() {
    	//TODO: config delim?
    	return this.getName() + ":" + this.getDistributorName();
    }

    public String getName() {
        return name;
    }
    
    public void setName(String name) {
		this.name = name;
	}

	public PipelinesDistributorBase getDistributor() {
		return distributor;
	}

	public void setDistributor(PipelinesDistributorBase distributor) {
		this.distributor = distributor;
	}

	public PipelineServiceConfig getConfig() {
		return config;
	}

	public void setConfig(PipelineServiceConfig config) {
		this.config = config;
	}

	public ArrayBlockingQueue<AbstractMessage> getInQ() {
        return inQ;
    }

	public void setInQ(ArrayBlockingQueue<AbstractMessage> inQ) {
		this.inQ = inQ;
	}

	public Counter getCounter() {
		return counter;
	}

	public void setCounter(Counter counter) {
		this.counter = counter;
	}

	public AbstractMessage getCurrMsg() {
		return currMsg;
	}

	public void setCurrMsg(AbstractMessage currMsg) {
		this.currMsg = currMsg;
	}

	public String getPipelineName() {
		return pipelineName;
	}

	public void setPipelineName(String pipelineName) {
		this.pipelineName = pipelineName;
	}
	
	public PipelinesMasterConfig getMasterConfig() {
		return this.getDistributor().getConfig().getMasterConfig();
	}
	
	public String getDistributorName() {
		return this.getConfig().getDistributorName();
	}
}
