package org.softwarepipelines.framework.springws;

import java.rmi.RemoteException;
import java.util.concurrent.ArrayBlockingQueue;

import javax.xml.rpc.ServiceException;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.PipelinesMasterConfig;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.util.ConfigHelper;
import org.softwarepipelines.framework.callback.PipelinesCallbackHandler;
import org.softwarepipelines.framework.distributor.PipelinesDistributorBase;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.util.PipelinesHelper;
import org.springframework.remoting.jaxrpc.ServletEndpointSupport;

public class PipelinesDistributorEndpoint
		extends ServletEndpointSupport implements PipelinesDistributorRemote, PipelinesCallbackHandler {
	private final static Logger LOG = Logger.getLogger(PipelinesDistributorEndpoint.class);
	
	// Pipelines Components
	protected PipelinesMasterConfig masterConfig;
	protected PipelinesDistributorBase distributor;
	protected PipelinesDistributorConfig distributorConfig;
	
	// response message queue
	protected ArrayBlockingQueue<AbstractMessage> responseQ;
	
	public PipelinesDistributorEndpoint() {
	}

    protected void onInit() throws ServiceException {
    	// get an instance of the master config bean
    	masterConfig = (PipelinesMasterConfig) getApplicationContext().getBean( "masterConfig" );
    	try {
    		masterConfig.load();
    	} catch (Exception e) {
    		throw new ServiceException("Error loading master config!", e);
    	}
        
        // get the distributor configuration bean, which should have a name specified ONLY
        distributorConfig = (PipelinesDistributorConfig) getApplicationContext().getBean( "distributorConfig" );
        
        try {
        	// get the distrbutor config from the master configuration, and finish the distrbutor config
            //distributor.configure(ConfigHelper.getDistributorConfigByName(masterConfig, distributorConfig.getName()));
            
            // get the distributor by name and start it
        	distributor = PipelinesHelper.getDistributorByName(distributorConfig.getName(), masterConfig);
        	distributor.start();
        } catch (PipelineException pe) {
        	throw new ServiceException(pe);
        }
        
        // TODO: instantiate repsonse queue
        responseQ = new ArrayBlockingQueue<AbstractMessage>(1);
    }

    // TODO: change to invokeSync and invokeAsync?
	public void invokeVoid(AbstractMessage msg) throws RemoteException {
    	LOG.info("Web service service invoke, distributing message: " + msg);
    	
    	// validate the message message
    	try {
    		validateMessage(msg, true);
    	} catch (PipelineException pe) {
    		throw new RemoteException("Error validating incoming message", pe);
    	}
    	
    	// and then distribute the message
    	try {
    		distributor.distribute(msg);
    	} catch (PipelineException pe) {
    		throw new RemoteException("Error distributing message", pe);
    	}
    }
	
	public AbstractMessage invokeReturn(AbstractMessage msg) throws RemoteException {
    	LOG.info("Web service service invoke, distributing message: " + msg);
    	
    	// validate the message
    	try {
    		validateMessage(msg, false);
    	} catch (PipelineException pe) {
    		throw new RemoteException("Error validating incoming message", pe);
    	}
    	
    	// then distribute the message
    	try {
    		distributor.distribute(msg);
    	} catch (PipelineException pe) {
    		throw new RemoteException("Error distributing message", pe);
    	}
    	
    	// finally, wait for the response and return it
    	try {
    		return this.waitForResponse();
    	} catch (PipelineException pe) {
    		throw new RemoteException("Error waiting for response message", pe);
    	}
    }
	
	private void validateMessage(final AbstractMessage msg, final boolean expOneWay) throws PipelineException {
		// validate that we have a connector reference
    	if (msg.getConnectorRef() == null) {
    		throw new PipelineException("No connector reference set on message: " + msg);
    	}
    	
    	// check the service name to validate that it is a void service
    	try {
			if (ConfigHelper.isDistributorServiceOneWay(getDistributor().getConfig(), msg.getServiceName()) != expOneWay) {
	    		throw new PipelineException(PipelineException.NOT_A_RETURN_SERVICE_MSG + "\n\tDistributor: " + this.getDistributor().getConfig().getName() +
	    				"\n\tService: " + msg.getServiceName());
	    	}
    	} catch (ConfigException ce) {
			throw new PipelineException("Error checking if service " + msg.getServiceName() + " is relay type", ce);
		}
		
		// set the callback reference
    	distributor.addConnectorToCallbackRef(msg.getConnectorRef().toString(), this);
	}
	
	protected AbstractMessage waitForResponse() throws PipelineException {
        AbstractMessage response = null;

        while (response == null) {
            try {
                response = this.getResponseQ().take();
            } catch (InterruptedException ie) {
                LOG.warn("Interrupted while waiting for response AbstractMessage", ie);
            }
        }

        return response;
    }
	
	public void returnResult(final AbstractMessage result) throws PipelineException {
        // place the incoming result AbstractMessage onto the incomingMsgs queue to wake up the invokeReturnService thread
        do {
            try {
                this.getResponseQ().put(result);
                LOG.info("Put result message on incoming message queue...");
                break;
            } catch (InterruptedException ie) {
                LOG.warn("Interrupted while enqueueing received result: " +
                    result, ie);
            }
        } while (!this.getResponseQ().contains(result));
    }

	public PipelinesMasterConfig getMasterConfig() {
		return masterConfig;
	}

	public void setMasterConfig(PipelinesMasterConfig masterConfig) {
		this.masterConfig = masterConfig;
	}

	public PipelinesDistributorBase getDistributor() {
		return distributor;
	}

	public void setDistributor(PipelinesDistributorBase distributor) {
		this.distributor = distributor;
	}

	public ArrayBlockingQueue<AbstractMessage> getResponseQ() {
		return responseQ;
	}

	public void setResponseQ(ArrayBlockingQueue<AbstractMessage> responseQ) {
		this.responseQ = responseQ;
	}
	
}
