package org.softwarepipelines.framework.util;

import org.apache.log4j.Logger;

import java.util.Map;
import java.util.HashMap;

/**
 * Instrumentation class for collecting runtime performance data.
 */
public class Instrumentor {

    /**
     * Logger for outputting instrumentation statistics.
     */
    private static final Logger logger = Logger.getLogger( Instrumentor.class );

    /**
     * Map of instrumentors.
     */
    private static final Map<String,Instrumentor> map = new HashMap<String, Instrumentor>();

    /**
     * Name of this instrumentor instance.
     */
    private String name;

    /**
     * Map of method invocation data.
     */
    private Map<String, MethodInvocation> methodMap = new HashMap<String, MethodInvocation>();

    /**
     * Map of named counters which can be used for tracking arbitrary statistics.
     */
    private Map<String, Counter> counterMap = new HashMap<String, Counter>();
    
    /**
     * Configurable log interval.
     */
    private static int logInterval;

    /**
     * Static initializer spawns a logging thread to output statistics.
     */
    private static Thread loggingThread;
    static {
        loggingThread = new Thread( new LoggingThread() );
        loggingThread.setDaemon(true);
        loggingThread.start();

    }
    /**
     * Static shutdown method stops the logging thread
     */
    public static void shutdown() {
    	LoggingThread.alive = false;
    }
    /**
     * Private constructor to force use of getInstrumentor() method to obtain
     * an instance.
     */
    private Instrumentor(String name) {
        this.name = name;
    }

    /**
     * Get an Instrumentor instance for the specified class.
     *
     * @param theClass
     * @return
     */
    public static synchronized Instrumentor getInstrumentor(Class theClass) {
        return getInstrumentor( theClass.getName() );
    }

    /**
     * Get an Instrumentor instance for the specified class name.
     * 
     * @param className
     * @return
     */
    public static synchronized Instrumentor getInstrumentor(String className) {
        Instrumentor ret = map.get( className );
        if (ret == null) {
            ret = new Instrumentor(className);
            map.put( className, ret );
        }
        return ret;
    }

    public synchronized void recordMethodInvocation(String methodName, long duration) {
        //TODO:
    }

    	/**
    	 * Get a named Counter instance.
    	 */
    public synchronized Counter getCounter(String name) {
        final String key = this.name + ":" + name;
        Counter ret = counterMap.get( key );
        if (ret == null) {
            ret = new Counter();
            counterMap.put( key, ret );
        }
        return ret;
    }
    
    /**
     * Return the log interval.
     */
    public static int getLogInterval() {
		return logInterval;
	}

    /**
     * Set the log interval.
     */
	public static void setLogInterval(int aLogInterval) {
		logInterval = aLogInterval;
	}

    class MethodInvocation {
        private long count;
        private long duration;
    }

    /**
     * Perform logging in a separate thread to minimize overhead. 
     */
    static class LoggingThread implements Runnable {

        /*private static final int LOG_PERIOD;
        static { 
        	if (logInterval > 0) {
        		LOG_PERIOD = logInterval;
        		logger.info("Log interval is " + logInterval + " ms");
        	} else {
        		LOG_PERIOD = 1000; // TODO: make configurable
        	}
        }*/
    	public static boolean alive = true;
        private float total;
        private float oneSec = 1000;

        LoggingThread() {
        }

        public void run() {
            while (alive) {
                try {
                    Thread.sleep(logInterval);//LOG_PERIOD);
                } catch (InterruptedException e) {
                    logger.warn( "Sleep interrupted", e );
                }
                for (Instrumentor instrumentor : Instrumentor.map.values()) {
                    for (Map.Entry<?,?> entry : instrumentor.counterMap.entrySet()) {
                        Counter counter = (Counter) entry.getValue();
                        total = counter.reset();
                        logger.info( entry.getKey() + " = " + total + "(" + ((float)(oneSec*total)/((float)logInterval)) + " TPS)");
                    }
                }
            }
        }
    }

}
