package org.softwarepipelines.framework.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.util.HashMap;

import org.apache.log4j.Logger;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;

/**
 * Helper class containing convenience methods for reading and writing messages to and from buffers.
 * 
 * @author cholm
 *
 */
public class MessageHelper {
	private static final Logger LOG = Logger.getLogger(MessageHelper.class);

	/*=================================================================================
	 * -----------------------  WRITE MESSAGE TO BUFFER -------------------------------
	 *=================================================================================*/
	/**
	 * Writes a messages contents (through the externalizable interface calls) into a ByteBuffer for transmission
	 * over a socket channel.
	 */
	public static synchronized ByteBuffer writeMessageToBuffer(AbstractMessage msg, ByteBuffer writeBuf) throws PipelineException {
		synchronized (writeBuf) {
			// mark the buffer's current position, in case we need to rewind and compact to prevent overflow
			writeBuf.mark();
			
			if (!writeServiceNameHdr(msg, writeBuf)) {
				writeBuf.reset();
				return writeBuf;
			}
			
			// and then write the message contents
			if (!writeMessage(msg, writeBuf)) {
				writeBuf.reset();
				return writeBuf;
			}
	    	return writeBuf;
		}
	}
	
	/**
	 * Writes the service name header.
	 * 
	 * The service name header is read before the message contents itself, as we need to know which class to create an
	 * instance of so that we call into the correct externalizable methods.
	 */
	private static boolean writeServiceNameHdr(AbstractMessage msg, ByteBuffer writeBuf) throws PipelineException {
		final char[] srvNameChars = msg.getServiceName().toCharArray();
		final int len = srvNameChars.length;
		
		// check that we have space to write the length and the service 
		if (writeBuf.remaining() < ((len*2) + 4)) {
			return false;
		}
    	writeBuf.putInt(len);
    	for (int i = 0; i < len; i++) {
    		writeBuf.putChar(srvNameChars[i]);
    	}
    	return true;
    }
	
	private static boolean writeMessage(AbstractMessage msg, ByteBuffer writeBuf) throws PipelineException {
    	
    	// build the byte array and object output streams
    	ByteArrayOutputStream baos = new ByteArrayOutputStream();
    	ObjectOutputStream oos;
    	try {
    		oos = new ObjectOutputStream(baos);
    	} catch (IOException ioe) {
        	throw new PipelineException("Error opening ObjectOutputStream on ByteArrayOutputStream", ioe);
        }
    	
    	// write the message to the object output stream
    	try {
    		msg.writeExternal(oos);
    		baos.flush();
	    	oos.flush();
    	} catch (IOException ioe) {
        	throw new PipelineException("Error writing message to ObjectOutput", ioe);
        }
    	final byte[] msgBytes = baos.toByteArray();
    	
    	// check the write buffer's capacity
    	if (writeBuf.remaining() < msgBytes.length + 4) {
    		return false;
    	}
    	
    	// then, write the message bytes
    	writeBuf.putInt(msgBytes.length);
    	writeBuf.put(msgBytes);
    	return true;
    }
	
	/*=================================================================================
	 * -----------------------  READ MESSAGE FROM BUFFER -------------------------------
	 *=================================================================================*/
	public static synchronized AbstractMessage readMessageFromBuffer(ByteBuffer readBuf, HashMap<String, Class<?>> serviceToTypeMap) throws PipelineException {
		AbstractMessage msg;
		
		// mark the read buffer in case we have an incomplete message at the end of a buffer
		readBuf.mark();
		
		// read the service name
		final String serviceName = readServiceNameHdr(readBuf);
		
		// if the service name came back null, we have an incompete read that we need to complete, return null
		if (serviceName == null) {
			return null;
		}
		
		// check that the service name maps to a valid type
		if (!serviceToTypeMap.containsKey(serviceName)) {
			LOG.info("readBuf pos: " + readBuf.position() + "\tlimit: " + readBuf.limit() + "\tcap: " + readBuf.capacity());
			throw new PipelineException("Couldn't find message class for service: " + serviceName + 
					" in map: " + serviceToTypeMap);
		}
		
		// create new instance of the Message class
		try {
			msg = (AbstractMessage) serviceToTypeMap.get(serviceName).newInstance();
		} catch (Exception e) {
			throw new PipelineException("Error instantiating new instance of class " + serviceToTypeMap.get(serviceName), e);
		}
		boolean success = readMessage(msg, readBuf);
		if (!success) {
			return null;
		} else {
			return msg;
		}
	}
	
	private static String readServiceNameHdr(ByteBuffer readBuf) throws PipelineException {
		try {
			// if we have no room for the int value, compact and return to complete the message read
			if (readBuf.remaining() < 4) {
				readBuf.compact();
				return null;
			}
			int len = readBuf.getInt();
			
			// XXX
			if (len == 0) {
				LOG.info("Service Name length is zero!");
			}
			
			// if the buffer does not have the whole service name, reset to the beginning of the message and return null
			if (len*2 > readBuf.remaining()) {
				readBuf.reset();
				readBuf.compact();
				return null;
			}
			char[] srvNameChars = new char[len];
			for (int i = 0; i < len; i++) {
				srvNameChars[i] = readBuf.getChar();
			}
			
			return String.valueOf(srvNameChars);
		} catch (BufferUnderflowException bue) {
			throw new PipelineException(bue);
		}
    }
	
	private static boolean readMessage(AbstractMessage msg, ByteBuffer readBuf) throws PipelineException {
		// check that we ahve enough room for the length spec
		if (readBuf.remaining() < 4) {
			readBuf.reset();
			readBuf.compact();
			return false;
		}
		int len = readBuf.getInt();
		
		// and then see if the buffer has the whole message 
		if (len > readBuf.remaining()) {
			readBuf.reset();
			readBuf.compact();
			return false;
		}
        byte[] msgBytes = new byte[len];
        readBuf.get(msgBytes, 0, len);
        
        // open a ByteArrayInputStream on the message
        ByteArrayInputStream bais = new ByteArrayInputStream(msgBytes);
        ObjectInput in;
        try {
        	in = new ObjectInputStream(bais);
        } catch (IOException ioe) {
        	throw new PipelineException("Error building ObjectInputStream on top of ByteArrayInputStream", ioe);
        }
		
        // and read the message from the input stream
		try {
			msg.readExternal(in);
		} catch (IOException ioe) {
        	throw new PipelineException("Error reading message from ObjectInput!", ioe);
        }
		return true;
    }
	
	
}
