package org.softwarepipelines.framework.util;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.PipelinesMasterConfig;
import org.softwarepipelines.config.connector.PipelinesConnectorConfig;
import org.softwarepipelines.config.dest.DestinationInfo;
import org.softwarepipelines.config.dest.DestinationMap;
import org.softwarepipelines.config.distributor.PipelinesDistributorConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.config.util.ConfigHelper;
import org.softwarepipelines.framework.connector.PipelinesConnector;
import org.softwarepipelines.framework.connector.PipelinesConnectorFactory;
import org.softwarepipelines.framework.distributor.PipelinesDistributorBase;
import org.softwarepipelines.framework.distributor.PipelinesDistributorFactory;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.nio.ClientMessageBroker;
import org.softwarepipelines.framework.nio.MessageBrokerFactory;
import org.springframework.context.ApplicationContext;

/**
 * Helper class with convenience methods for getting pipelines framework components.
 * 
 * @author cholm
 */
public class PipelinesHelper {
	private static final Logger LOG = Logger.getLogger(PipelinesHelper.class);
	
	/**
	 * Build and return a new <code>PipelinesDistributor</code> instance, configured in the given master config with the given name.
	 * 
	 * @param name the name of the distributor we wish to instantiate
	 * @param masterConfig the master pipelines configuration
	 * @return the new <code>PipelinesDistributor</code>
	 * @throws PipelineException
	 */
	public static PipelinesDistributorBase getDistributorByName(final String name, final PipelinesMasterConfig masterConfig) throws PipelineException {
		
		return PipelinesDistributorFactory.getPipelinesDistributor(ConfigHelper.getDistributorConfigByName(masterConfig, name));
	}
	
	/**
	 * 
	 * Build and return a new <code>PipelinesConnector</code> instance, configured by name on the given distributor in the master config.
	 * 
	 * @param distName the name of the <code>PipelinesDistributor</code> on which the desired <code>PipelinesConnector</code> is configured.
	 * @param connName the name of the <code>PipelinesConnector</code>
	 * @param masterConfig the master pipelines framework configuration in which both the distributor and connector are configured
	 * @return a <code>PipelinesConnector</code>
	 * @throws PipelineException if no such connector or distributor is configured, or something else goes wrong
	 */
	public static PipelinesConnector getConnectorByName(final String distName, final String connName, final PipelinesMasterConfig masterConfig) throws PipelineException {
		PipelinesDistributorConfig distConfig;
		PipelinesConnectorConfig connConfig;
		
		// get the distributor by name
		distConfig = ConfigHelper.getDistributorConfigByName(masterConfig, distName);
		
		// and then the connector config
		try {
			connConfig = ConfigHelper.getConnectorConfigByName(distConfig, connName);
		} catch (ConfigException ce) {
			throw new PipelineException("Error getting config for connector '" + connName + "' on distributor '" + distName + "'", ce);
		}
		
		// XXX if we are configuring a local connector for a service on the same distributor, if we try to create the distributor instance
		// again here we get stuck in an infinite loop
		
		// if we have a local type connector, get the distributor instance and checkout the connector
		if (connConfig.getType().equals(PipelinesConnectorConfig.LOCAL_CONNECTOR_TYPE)) {
			// get the distributor from the factory
			PipelinesDistributorBase distributor = PipelinesDistributorFactory.getPipelinesDistributor(distConfig);
			
			// checkout a local connector from the distributor and return it
			return distributor.checkoutLocalConnector();
		}
		// else if we have a socket connector, get the client message broker from the factory
		else if (connConfig.getType().equals(PipelinesConnectorConfig.SOCKET_CONNECTOR_TYPE)) {
			ClientMessageBroker clientBroker = MessageBrokerFactory.getClientMessageBroker(distConfig);
			
			// start the broker
			clientBroker.start();
			
			// and checkout and return a socket connector
			return clientBroker.checkoutSocketConnector();
		}
		// else we have a web service connector, build it using the factory
		else {
			return PipelinesConnectorFactory.getPipelinesConnector(connConfig);
		}
	}
	
	/**
	 * Get a <code>PipelinesConnector</code> from a <code>DestinationMap</code> configured in a Spring ApplicationContext.
	 * 
	 * @param ac the ApplicationContext
	 * @param baseKey the base map key, attained in a service through the <code>PipelineServiceBase.getDestMapKey()</code> method.
	 * @param targetService the name of the target service
	 * @return the <code>PipelinesConnector</code>
	 * @throws PipelineException
	 */
	public static PipelinesConnector getConnectorFromDestMap(final ApplicationContext ac, 
			final String baseKey, final String targetService) throws PipelineException {
		DestinationMap map;
		DestinationInfo destInfo;
		PipelinesDistributorConfig distConfig;
		PipelinesConnectorConfig connConfig;
		String key;
		
		// try to load the map
		try {
			map = (DestinationMap) ac.getBean(DestinationMap.DEST_MAP_BEAN_NAME);
		} catch (Exception e) {
			throw new PipelineException("Error loading DestinationMap bean");
		}
		
		// create the full key
		key = baseKey + ":" + targetService;
		
		// validate that the map contains the destination info, and get it
		if (!map.getDestinationMap().containsKey(key)) {
			throw new PipelineException("No key '" + key + "' in DestinationMap");
		}
		destInfo = map.getDestinationMap().get(key);
		
		// validate the destination info and destination map
		if (map.getMasterConfig() == null || destInfo.getDestinationDistributor() == null || destInfo.getDestinationConnector() == null) {
			throw new PipelineException("Invalid DestinationInfo!  Must include a non-null references to a PipelinesMasterConfig, and " +
					" the name of the destination PipelinesDistributor and PipelinesConnector in that config.");
		}
		
		// get the distributor by name
		distConfig = ConfigHelper.getDistributorConfigByName(map.getMasterConfig(), destInfo.getDestinationDistributor());
		
		// and then the connector config
		try {
			connConfig = ConfigHelper.getConnectorConfigByName(distConfig, destInfo.getDestinationConnector());
		} catch (ConfigException ce) {
			throw new PipelineException("Error getting config for connector '" + destInfo.getDestinationConnector() + "' on distributor '" + 
					distConfig.getName() + "'", ce);
		}
		
		// if we have a local type connector, get the distributor instance and checkout the connector
		LOG.debug("Helper retrieving connector " + connConfig.getName() + " on distributor " + distConfig.getName());
		if (connConfig.getType().equals(PipelinesConnectorConfig.LOCAL_CONNECTOR_TYPE)) {
			
			// get the distributor from the factory
			PipelinesDistributorBase distributor = PipelinesDistributorFactory.getPipelinesDistributor(distConfig);
			
			// checkout a local connector from the distributor and return it
			return distributor.checkoutLocalConnector();
		}
		// else if we have a socket connector, get the client message broker from the factory
		else if (connConfig.getType().equals(PipelinesConnectorConfig.SOCKET_CONNECTOR_TYPE)) {
			ClientMessageBroker clientBroker = MessageBrokerFactory.getClientMessageBroker(distConfig);
			
			// start the broker
			clientBroker.start();
			
			// and checkout and return a socket connector
			return clientBroker.checkoutSocketConnector();
		}
		// else we have a web service connector, build it using the factory
		else {
			return PipelinesConnectorFactory.getPipelinesConnector(connConfig);
		}
	}
	
}
