package org.softwarepipelines.framework.connector;

import java.io.FileNotFoundException;
import java.io.IOException;

import junit.framework.TestCase;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.PipelinesMasterConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.examples.hello.HelloMessage;
import org.softwarepipelines.examples.hello.id.HelloIdMessage;
import org.softwarepipelines.framework.distributor.PipelinesDistributorBase;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.util.PipelinesHelper;
import org.softwarepipelines.framework.util.PipelinesTestHelper;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;


public class LocalPipelinesConnectorTestCase extends TestCase {
    private static final Logger LOG = Logger.getLogger(LocalPipelinesConnectorTestCase.class);
    
    // Application Context
    private ApplicationContext ac;
    
    // Framework components
    private PipelinesConnector connector;
    private PipelinesDistributorBase distributor;
    private PipelinesMasterConfig masterConfig;
    
    public LocalPipelinesConnectorTestCase(String name) {
        super(name);
        
        // load the framework test application context
        ac = new ClassPathXmlApplicationContext(PipelinesTestHelper.FRAMEWORK_TEST_CONTEXT_FILE);
    }

    public void tearDown() throws PipelineException {
        // shut down the distributor and the connector
    	connector.shutdown();
        distributor.shutdown();
    }
    
    private void setupDistributor(final String beanName) throws PipelineException, ConfigException, FileNotFoundException, IOException {
    	// setup up the distributor
    	masterConfig = (PipelinesMasterConfig) ac.getBean(beanName);
        distributor = PipelinesHelper.getDistributorByName("localDistributor", masterConfig);
        distributor.start();
        
        // and checkout a local connector
        connector = PipelinesHelper.getConnectorByName(distributor.getConfig().getName(), "localConnector", masterConfig);
    }

    public void testContentBasedLocalVoidInvoke() throws Exception {
    	setupDistributor("singleTierCBConfig");
    	
        // test that we can successfully invoke a void pipeline service, alpha routing by matching reg exp
    	HelloMessage msg = new HelloMessage();
    	msg.setServiceName("helloService");
    	msg.setName("Andy");
        connector.send(msg);
        
        // and then that we can successfully invoke a void pipeline service routing by integer range
        HelloIdMessage idMsg = new HelloIdMessage();
        idMsg.setServiceName("helloIdService");
        idMsg.setId(34);
        connector.send(idMsg);
    }

    public void testContentBasedLocalReturnInvoke() throws Exception {
    	AbstractMessage resp;
    	setupDistributor("singleTierCBConfig");
    	
    	// build the message with a void destination service
    	HelloMessage msg = new HelloMessage();
    	msg.setServiceName("helloService");
    	msg.setName("Jeff");
    	
    	// and test that we throw the proper exception
    	try {
    		resp = connector.sendAndReceive(msg);
    	} catch (PipelineException pe) {
    		assertEquals(PipelineException.NOT_A_RETURN_SERVICE_MSG + 
    				"\n\tDistributor: " + distributor.getConfig().getName() + 
    				"\n\tService: " + msg.getServiceName(), 
    				pe.getMessage());
    	} catch (Exception e) {
    		fail("Should have caught a PipelineException! Caught: " + e);
    	}
    	
        // then set the return destination service
    	msg.setServiceName("helloRetService");
    	
    	// and test that we get a valid response
    	resp = connector.sendAndReceive(msg);
        assertNotNull(resp);
        assertEquals(msg.getServiceName(), resp.getServiceName());
        LOG.info("Got response: " + resp);
    }
    
    public void testRoundRobinLocalVoidInvoke() throws Exception {
    	setupDistributor("singleTierRRConfig");
    	
        // test that we can successfully invoke a void pipeline service
    	HelloMessage msg = new HelloMessage();
    	msg.setServiceName("helloService");
    	msg.setName("Andy");
        connector.send(msg);
    }

    public void testRoundRobinLocalReturnInvoke() throws Exception {
    	AbstractMessage resp;
    	setupDistributor("singleTierRRConfig");
    	
        // build the message with the void destination service
    	HelloMessage msg = new HelloMessage();
    	msg.setServiceName("helloService");
    	msg.setName("Jeff");
    	
    	// and check that we throw the appropriate pipeline exception
    	// TODO: plug exception messages into spring/properties file for i18n and testing purposes
    	try {
    		resp = connector.sendAndReceive(msg);
    	} catch (PipelineException pe) {
    		assertEquals(PipelineException.NOT_A_RETURN_SERVICE_MSG + 
    				"\n\tDistributor: " + distributor.getConfig().getName() + 
    				"\n\tService: " + msg.getServiceName(), 
    				pe.getMessage());
    	} catch (Exception e) {
    		fail("Should have thrown a PipelineException! Caught: " + e);
    	}
    	
    	// then set the return service name
    	msg.setServiceName("helloRetService");
    	
    	// and test that we can invoke the return service
        resp = connector.sendAndReceive(msg);
        assertNotNull(resp);
        assertEquals(msg.getServiceName(), resp.getServiceName());
        LOG.info("Got response: " + resp);
    }
}
