package org.softwarepipelines.framework.distributor;

import java.io.FileNotFoundException;
import java.io.IOException;

import org.apache.log4j.Logger;
import org.softwarepipelines.config.PipelinesMasterConfig;
import org.softwarepipelines.config.exception.ConfigException;
import org.softwarepipelines.examples.hello.HelloMessage;
import org.softwarepipelines.framework.connector.PipelinesConnector;
import org.softwarepipelines.framework.exception.PipelineException;
import org.softwarepipelines.framework.message.AbstractMessage;
import org.softwarepipelines.framework.util.PipelinesHelper;
import org.softwarepipelines.framework.util.PipelinesTestHelper;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import junit.framework.TestCase;

public class SingleTierDistributorTestCase extends TestCase {
	private static final Logger LOG = Logger.getLogger(SingleTierDistributorTestCase.class);
	
	/* Framework Test Application Context */
	private ApplicationContext ac;
	
	/* Pipelines Framework Component */
    private PipelinesMasterConfig masterConfig;
    private PipelinesDistributorBase distributor;
    private PipelinesConnector localConn, socketConn;
	
	public SingleTierDistributorTestCase(final String name) {
		super(name);
		
		ac = new ClassPathXmlApplicationContext(PipelinesTestHelper.FRAMEWORK_TEST_CONTEXT_FILE);
	}
	
	public void setUp() {	
	}
	
	public void tearDown() throws PipelineException {
		 // shut down the distributor in the invoker
		localConn.shutdown();
		socketConn.shutdown();
		distributor.shutdown();
	}
	
	private void setupDistributor(final String beanName) throws PipelineException, ConfigException, FileNotFoundException, IOException {
    	// setup up the service invoker
    	masterConfig = (PipelinesMasterConfig) ac.getBean(beanName);
    	
    	// get the distributor
        distributor = PipelinesHelper.getDistributorByName("localDistributor", masterConfig);
        distributor.start();
        
        // checkout the connectors
        socketConn = PipelinesHelper.getConnectorByName(distributor.getConfig().getName(), "socketConnector", masterConfig);
        localConn = PipelinesHelper.getConnectorByName(distributor.getConfig().getName(), "localConnector", masterConfig);
    }
	
	public void testContentBasedVoidInvoke() throws Exception {
    	setupDistributor("singleTierCBConfig");
    	
        // test the void invokes
    	for (int i = 0; i < PipelinesTestHelper.helloVoidMsgs.length; i++) {
    		localConn.send(PipelinesTestHelper.helloVoidMsgs[i]);
    	}
    	for (int i = 0; i < PipelinesTestHelper.helloVoidMsgs.length; i++) {
    		socketConn.send(PipelinesTestHelper.helloVoidMsgs[i]);
    	}
        
        // wait for the hello service to be enqueued before we call tear down and shut the broker and distrbutor down
        try {Thread.sleep(10);} catch (Exception e) {}
    }

    public void testContentBasedReturnInvoke() throws Exception {
    	AbstractMessage response;
    	HelloMessage msg;
    	
    	setupDistributor("singleTierCBConfig");
    	
        // test the return invokes with both connectors
    	msg = PipelinesTestHelper.helloRetMsgs[0];
        response = localConn.sendAndReceive(msg);
        assertNotNull(response);
        assertEquals(msg.getServiceName(), response.getServiceName());
        LOG.info("Local Connector got response: " + response);
        
        msg = PipelinesTestHelper.helloRetMsgs[1];
        response = socketConn.sendAndReceive(msg);
        assertNotNull(response);
        assertEquals(msg.getServiceName(), response.getServiceName());
        LOG.info("Got response: " + response);
    }
    
    public void testRoundRobinVoidInvoke() throws Exception {
    	setupDistributor("singleTierRRConfig");
    	
        // test the void invokes
    	for (int i = 0; i < PipelinesTestHelper.helloVoidMsgs.length; i++) {
    		localConn.send(PipelinesTestHelper.helloVoidMsgs[i]);
    	}
    	for (int i = 0; i < PipelinesTestHelper.helloVoidMsgs.length; i++) {
    		socketConn.send(PipelinesTestHelper.helloVoidMsgs[i]);
    	}
    }

    public void testRoundRobinReturnInvoke() throws Exception {
    	AbstractMessage response = null;
    	HelloMessage msg;
    	long start, end;
    	
    	setupDistributor("singleTierRRConfig");
    	
        // test the return invokes with both connectors
    	msg = PipelinesTestHelper.helloRetMsgs[2];
    	start = System.currentTimeMillis();
        for (int i = 0; i < 20; i++) {
	        response = localConn.sendAndReceive(msg);
	        assertNotNull(response);
	        LOG.debug("Got response: " + response);
        }
        end = System.currentTimeMillis();
        try {Thread.sleep(100);} catch (Exception e) {}
        LOG.info("Invoking helloRetService twenty times from a single client thread through a local connector should take about 1000 ms, took " + 
        		(end - start) + " ms");
        assertNotNull(response);
        assertEquals(msg.getServiceName(), response.getServiceName());
        LOG.info("Local Connector got response: " + response);
        
        // time sending twenty messages, should be close to one fourth of a sec given that we have 4 rr pipelines
        msg = PipelinesTestHelper.helloRetMsgs[3];
        start = System.currentTimeMillis();
        for (int i = 0; i < 20; i++) {
	        response = socketConn.sendAndReceive(msg);
	        LOG.debug("Got response: " + response);
        }
        end = System.currentTimeMillis();
        try {Thread.sleep(100);} catch (Exception e) {}
        LOG.info("Invoking helloRetService twenty times from a single client thread should take about 1000 ms, took " + (end - start) + " ms");
        assertNotNull(response);
        assertEquals(msg.getServiceName(), response.getServiceName());
        LOG.info("Got response: " + response);
    }
}
